# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import logging.config
import os
import sys


class _TaipyLogger:
    _ENVIRONMENT_VARIABLE_NAME_WITH_LOGGER_CONFIG_PATH = "TAIPY_LOGGER_CONFIG_PATH"

    __logger = None

    @classmethod
    def _get_logger(cls):
        cls._ENVIRONMENT_VARIABLE_NAME_WITH_LOGGER_CONFIG_PATH = "TAIPY_LOGGER_CONFIG_PATH"

        if cls.__logger:
            return cls.__logger

        if config_filename := os.environ.get(cls._ENVIRONMENT_VARIABLE_NAME_WITH_LOGGER_CONFIG_PATH):
            logging.config.fileConfig(config_filename)
            cls.__logger = logging.getLogger("Taipy")
        else:
            cls.__logger = logging.getLogger("Taipy")
            cls.__logger.setLevel(logging.INFO)
            ch = logging.StreamHandler(sys.stdout)
            ch.setLevel(logging.INFO)
            f = logging.Formatter("[%(asctime)s.%(msecs)03d][%(name)s][%(levelname)s] %(message)s", "%Y-%m-%d %H:%M:%S")
            ch.setFormatter(f)
            cls.__logger.addHandler(ch)
        return cls.__logger



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import sys
from abc import abstractmethod
from difflib import SequenceMatcher
from typing import List, Optional

from taipy.common.logger._taipy_logger import _TaipyLogger

from ._taipy_parser import _TaipyParser


class _AbstractCLI:
    _logger = _TaipyLogger._get_logger()

    _COMMAND_NAME: Optional[str] = None
    _ARGUMENTS: List[str] = []
    SIMILARITY_THRESHOLD = 0.8

    @classmethod
    @abstractmethod
    def create_parser(cls):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def handle_command(cls):
        raise NotImplementedError

    @classmethod
    def _parse_arguments(cls):
        args, unknown_args = _TaipyParser._parser.parse_known_args()

        if getattr(args, "which", None) != cls._COMMAND_NAME:
            return

        if unknown_args:
            _TaipyParser._sub_taipyparsers.get(cls._COMMAND_NAME).print_help()
            for unknown_arg in unknown_args:
                if not unknown_arg.startswith("-"):
                    continue

                error_message = f"Unknown arguments: {unknown_arg}."
                similar_args = [
                    arg
                    for arg in cls._ARGUMENTS
                    if SequenceMatcher(None, unknown_arg, arg).ratio() > cls.SIMILARITY_THRESHOLD
                ]
                if similar_args:
                    error_message += f" Did you mean: {' or '.join(similar_args)}?"
                cls._logger.error(error_message)
            cls._logger.error("Please refer to the help message above for more information.")
            sys.exit(1)

        return args



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import json
from datetime import timedelta
from typing import Any, Callable, Dict, List, Optional, Union

from taipy.common.config._config import _Config
from taipy.core.config import CoreSection, DataNodeConfig, JobConfig, ScenarioConfig, TaskConfig

from .checker.issue_collector import IssueCollector
from .common._classproperty import _Classproperty
from .common._config_blocker import _ConfigBlocker
from .common.frequency import Frequency
from .common.scope import Scope
from .global_app.global_app_config import GlobalAppConfig
from .section import Section
from .unique_section import UniqueSection

class Config:
    """Singleton class that manages the configuration of a Taipy application.

    The `Config` singleton is the main class to use for configuring a Taipy application.
    In particular, this class provides:

    1. Various methods to configure the application's behavior

        The Config class provides various methods to configure the application. Each method adds
        a specific section to the configuration and returns it.

        ??? example "Most frequently used configuration methods"

            ```python
            from taipy import Config

            def by_two(x: int):
                return x * 2

            input_cfg = Config.configure_data_node("my_input")
            result_cfg = Config.configure_data_node("my_result")
            task_cfg = Config.configure_task("my_double", function=by_two, input=input_cfg, output=result_cfg)
            scenario_cfg = Config.configure_scenario("my_scenario", task_configs=[task_cfg])

            Config.load("config.toml") # Load a configuration file
            ```

        ??? example "Advanced use case"

            The configuration can be done in three ways: Python code, configuration files, or
            environment variables. All configuration manners are ultimately merged (overriding the previous
            way) to create a final applied configuration. Please refer to the
            [advanced configuration](../../../../../../userman/advanced_features/configuration/advanced-config.md)
            section from the user manual for more details.

    2. Attributes and methods to retrieve the configuration values.

        Once the configuration is done, you can retrieve the configuration values using the exposed
        attributes.

        ??? Example "Retrieve configuration values"

            ```python
            from taipy import Config

            global_cfg = Config.global_config  # Retrieve the global application configuration
            data_node_cfgs = Config.data_nodes  # Retrieve all data node configurations
            scenario_cfgs = Config.scenarios  # Retrieve all scenario configurations
            ```

    3. A few methods to manage the configuration:

        The Config class also provides a few methods to manage the configuration.

        ??? example "Manage the configuration"

            - *Check the configuration for issues*: Use the `Config.check()^` method to check the
                configuration. It returns an `IssueCollector^` containing all the
                 `Issue^`s found. The issues are logged to the console for debugging.
            - *Block the configuration update*: Use the `Config.block_update()^` method to forbid
                any update on the configuration. This can be useful when you want to ensure that
                the configuration is not modified at run time. Note that running the `Orchestrator^`
                service` automatically blocks the configuration update.
            - *Unblock the configuration update*: Use the `Config.unblock_update()^` method to allow
                again the update on the configuration.
            - *Backup the configuration*: Use the `Config.backup()^` method to back up as a TOML
                file the applied configuration. The applied configuration backed up is the result
                of the compilation of the three possible configuration methods that overrides each
                others.
            - *Restore the configuration*: Use the `Config.restore()^` method to restore a TOML
                configuration file and replace the current applied configuration.
            - *Export the configuration*: Use the `Config.export()^` method to export as a TOML file
                the Python code configuration.
            - *Load the configuration*: Use the `Config.load()^` method to load a TOML configuration
                file and replace the current Python configuration.
            - *Override the configuration*: Use the `Config.override()^` method to load a TOML
                configuration file and override the current Python configuration.
    """
    @_Classproperty
    def unique_sections(cls) -> Dict[str, UniqueSection]:
        """A dictionary containing all unique sections."""

    @_Classproperty
    def sections(cls) -> Dict[str, Dict[str, Section]]:
        """A dictionary containing all non-unique sections."""

    @_Classproperty
    def global_config(cls) -> GlobalAppConfig:
        """configuration values related to the global application as a `GlobalAppConfig^`."""

    @classmethod
    @_ConfigBlocker._check()
    def load(cls, filename: str) -> None:
        """Load a configuration file.

        The current Python configuration is replaced and the Config compilation is triggered.

        Parameters:
            filename (Union[str, Path]): The path of the toml configuration file to load.
        """

    @classmethod
    def export(cls, filename: str) -> None:
        """Export a configuration.

        The export is done in a toml file. The exported configuration is taken
        from the Python code configuration.

        Note:
            If *filename* already exists, it is overwritten.

        Parameters:
            filename (Union[str, Path]): The path of the file to export.
        """

    @classmethod
    def backup(cls, filename: str) -> None:
        """Backup a configuration.

        The backup is done in a toml file.

        The backed up configuration is a compilation from the three possible methods to configure
        the application: the Python code configuration, the file configuration and the environment
        configuration.

        Note:
            If *filename* already exists, it is overwritten.

        Parameters:
            filename (Union[str, Path]): The path of the file to export.
        """

    @classmethod
    @_ConfigBlocker._check()
    def restore(cls, filename: str) -> None:
        """Restore a configuration file and replace the current applied configuration.

        Parameters:
            filename (Union[str, Path]): The path of the toml configuration file to load.
        """

    @classmethod
    @_ConfigBlocker._check()
    def override(cls, filename: str) -> None:
        """Load a configuration from a file and overrides the current config.

        Parameters:
            filename (Union[str, Path]): The path of the toml configuration file to load.
        """

    @classmethod
    def block_update(cls) -> None:
        """Block update on the configuration signgleton."""

    @classmethod
    def unblock_update(cls) -> None:
        """Unblock update on the configuration signgleton."""

    @classmethod
    @_ConfigBlocker._check()
    def configure_global_app(cls, **properties) -> GlobalAppConfig:
        """Configure the global application.

        Parameters:
            **properties (Dict[str, Any]): A dictionary of additional properties.

        Returns:
            The global application configuration.
        """

    @classmethod
    def check(cls) -> IssueCollector:
        """Check configuration.

        This method logs issue messages and returns an issue collector.

        Returns:
            Collector containing the info, warning and error issues.

        Raises:
            SystemExit: If configuration errors are found, the application
                exits with an error message.
        """

    @classmethod
    @_ConfigBlocker._check()
    def _register_default(cls, default_section: Section) -> None:
        """"""

    @classmethod
    @_ConfigBlocker._check()
    def _register(cls, section) -> None:
        """"""

    @classmethod
    def _override_env_file(cls) -> None:
        """"""

    @classmethod
    def _compile_configs(cls) -> None:
        """"""

    @classmethod
    def _to_json(cls, _config: _Config) -> str:
        """"""

    @classmethod
    def _from_json(cls, config_as_str: str) -> _Config:
        """"""

    @_Classproperty
    def job_config(cls) -> JobConfig:
        """"""

    @_Classproperty
    def data_nodes(cls) -> Dict[str, DataNodeConfig]:
        """"""

    @_Classproperty
    def tasks(cls) -> Dict[str, TaskConfig]:
        """"""

    @_Classproperty
    def scenarios(cls) -> Dict[str, ScenarioConfig]:
        """"""

    @_Classproperty
    def core(cls) -> Dict[str, CoreSection]:
        """"""

    @staticmethod
    def configure_scenario(
        id: str,
        task_configs: Optional[List[TaskConfig]] = None,
        additional_data_node_configs: Optional[List[DataNodeConfig]] = None,
        frequency: Optional[Frequency] = None,
        comparators: Optional[Dict[str, Union[List[Callable], Callable]]] = None,
        sequences: Optional[Dict[str, List[TaskConfig]]] = None,
        **properties,
    ) -> "ScenarioConfig":
        """Configure a new scenario configuration.

        Parameters:
            id (str): The unique identifier of the new scenario configuration.
            task_configs (Optional[List[TaskConfig^]]): The list of task configurations used by this
                scenario configuration. The default value is None.
            additional_data_node_configs (Optional[List[DataNodeConfig^]]): The list of additional data nodes
                related to this scenario configuration. The default value is None.
            frequency (Optional[Frequency^]): The scenario frequency.<br/>
                It corresponds to the recurrence of the scenarios instantiated from this
                configuration. Based on this frequency each scenario will be attached to the
                relevant cycle.
            comparators (Optional[Dict[str, Union[List[Callable], Callable]]]): The list of
                functions used to compare scenarios. A comparator function is attached to a
                scenario's data node configuration. The key of the dictionary parameter
                corresponds to the data node configuration id. During the scenarios'
                comparison, each comparator is applied to all the data nodes instantiated from
                the data node configuration attached to the comparator. See
                `(taipy.)compare_scenarios()^` more more details.
            sequences (Optional[Dict[str, List[TaskConfig]]]): Dictionary of sequence descriptions.
                The default value is None.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new scenario configuration.
        """

    @staticmethod
    def set_default_scenario_configuration(
        task_configs: Optional[List[TaskConfig]] = None,
        additional_data_node_configs: List[DataNodeConfig] = None,
        frequency: Optional[Frequency] = None,
        comparators: Optional[Dict[str, Union[List[Callable], Callable]]] = None,
        sequences: Optional[Dict[str, List[TaskConfig]]] = None,
        **properties,
    ) -> "ScenarioConfig":
        """Set the default values for scenario configurations.

        This function creates the *default scenario configuration* object,
        where all scenario configuration objects will find their default
        values when needed.

        Parameters:
            task_configs (Optional[List[TaskConfig^]]): The list of task configurations used by this
                scenario configuration.
            additional_data_node_configs (Optional[List[DataNodeConfig^]]): The list of additional data nodes
                related to this scenario configuration.
            frequency (Optional[Frequency^]): The scenario frequency.
                It corresponds to the recurrence of the scenarios instantiated from this
                configuration. Based on this frequency each scenario will be attached to
                the relevant cycle.
            comparators (Optional[Dict[str, Union[List[Callable], Callable]]]): The list of
                functions used to compare scenarios. A comparator function is attached to a
                scenario's data node configuration. The key of the dictionary parameter
                corresponds to the data node configuration id. During the scenarios'
                comparison, each comparator is applied to all the data nodes instantiated from
                the data node configuration attached to the comparator. See
                `taipy.compare_scenarios()^` more more details.
            sequences (Optional[Dict[str, List[TaskConfig]]]): Dictionary of sequences. The default value is None.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new default scenario configuration.
        """

    @staticmethod
    def set_default_data_node_configuration(
        storage_type: str, scope: Optional[Scope] = None, validity_period: Optional[timedelta] = None, **properties
    ) -> "DataNodeConfig":
        """Set the default values for data node configurations.

        This function creates the _default data node configuration_ object,
        where all data node configuration objects will find their default
        values when needed.

        Parameters:
            storage_type (str): The default storage type for all data node configurations.
                The possible values are *"pickle"* (the default value), *"csv"*, *"excel"*,
                *"sql"*, *"mongo_collection"*, *"in_memory"*, *"json"*, *"parquet"*, *"generic"*,
                or *"s3_object"*.
            scope (Optional[Scope^]): The default scope for all data node configurations.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The default data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_data_node_from(
        cls,
        source_configuration: "DataNodeConfig",
        id: str,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new data node configuration from an existing one.

        Parameters:
            source_configuration (DataNodeConfig): The source data node configuration.
            id (str): The unique identifier of the new data node configuration.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.<br/>
                The default properties are the properties of the source data node configuration.

        Returns:
            The new data node configuration.
        """

    @classmethod
    def configure_data_node(
        cls,
        id: str,
        storage_type: Optional[str] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new data node configuration.

        Parameters:
            id (str): The unique identifier of the new data node configuration.
            storage_type (Optional[str]): The data node configuration storage type. The possible values
                are None (which is the default value of *"pickle"*, unless it has been overloaded by the
                *storage_type* value set in the default data node configuration
                (see `(Config.)set_default_data_node_configuration()^`)), *"pickle"*, *"csv"*, *"excel"*,
                *"sql_table"*, *"sql"*, *"json"*, *"parquet"*, *"mongo_collection"*, *"in_memory"*, or
                *"generic"*.
            scope (Optional[Scope^]): The scope of the data node configuration.<br/>
                The default value is `Scope.SCENARIO` (or the one specified in
                `(Config.)set_default_data_node_configuration()^`).
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_csv_data_node(
        cls,
        id: str,
        default_path: Optional[str] = None,
        encoding: Optional[str] = None,
        has_header: Optional[bool] = None,
        exposed_type: Optional[str] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new CSV data node configuration.

        Parameters:
            id (str): The unique identifier of the new CSV data node configuration.
            default_path (Optional[str]): The default path of the CSV file.
            encoding (Optional[str]): The encoding of the CSV file.
            has_header (Optional[bool]): If True, indicates that the CSV file has a header.
            exposed_type (Optional[str]): The exposed type of the data read from CSV file.<br/>
                The default value is `pandas`.
            scope (Optional[Scope^]): The scope of the CSV data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new CSV data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_json_data_node(
        cls,
        id: str,
        default_path: Optional[str] = None,
        encoding: Optional[str] = None,
        encoder: Optional[json.JSONEncoder] = None,
        decoder: Optional[json.JSONDecoder] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new JSON data node configuration.

        Parameters:
            id (str): The unique identifier of the new JSON data node configuration.
            default_path (Optional[str]): The default path of the JSON file.
            encoding (Optional[str]): The encoding of the JSON file.
            encoder (Optional[json.JSONEncoder]): The JSON encoder used to write data into the JSON file.
            decoder (Optional[json.JSONDecoder]): The JSON decoder used to read data from the JSON file.
            scope (Optional[Scope^]): The scope of the JSON data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new JSON data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_parquet_data_node(
        cls,
        id: str,
        default_path: Optional[str] = None,
        engine: Optional[str] = None,
        compression: Optional[str] = None,
        read_kwargs: Optional[Dict] = None,
        write_kwargs: Optional[Dict] = None,
        exposed_type: Optional[str] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new Parquet data node configuration.

        Parameters:
            id (str): The unique identifier of the new Parquet data node configuration.
            default_path (Optional[str]): The default path of the Parquet file.
            engine (Optional[str]): Parquet library to use. Possible values are *"fastparquet"* or
                *"pyarrow"*.<br/>
                The default value is *"pyarrow"*.
            compression (Optional[str]): Name of the compression to use. Possible values are *"snappy"*,
                *"gzip"*, *"brotli"*, or *"none"* (no compression). The default value is *"snappy"*.
            read_kwargs (Optional[dict]): Additional parameters passed to the `pandas.read_parquet()`
                function.
            write_kwargs (Optional[dict]): Additional parameters passed to the
                `pandas.DataFrame.write_parquet()` function.<br/>
                The parameters in *read_kwargs* and *write_kwargs* have a **higher precedence** than the
                top-level parameters which are also passed to Pandas.
            exposed_type (Optional[str]): The exposed type of the data read from Parquet file.<br/>
                The default value is `pandas`.
            scope (Optional[Scope^]): The scope of the Parquet data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new Parquet data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_excel_data_node(
        cls,
        id: str,
        default_path: Optional[str] = None,
        has_header: Optional[bool] = None,
        sheet_name: Optional[Union[List[str], str]] = None,
        exposed_type: Optional[str] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new Excel data node configuration.

        Parameters:
            id (str): The unique identifier of the new Excel data node configuration.
            default_path (Optional[str]): The path of the Excel file.
            has_header (Optional[bool]): If True, indicates that the Excel file has a header.
            sheet_name (Optional[Union[List[str], str]]): The list of sheet names to be used.
                This can be a unique name.
            exposed_type (Optional[str]): The exposed type of the data read from Excel file.<br/>
                The default value is `pandas`.
            scope (Optional[Scope^]): The scope of the Excel data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new Excel data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_generic_data_node(
        cls,
        id: str,
        read_fct: Optional[Callable] = None,
        write_fct: Optional[Callable] = None,
        read_fct_args: Optional[List] = None,
        write_fct_args: Optional[List] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new generic data node configuration.

        Parameters:
            id (str): The unique identifier of the new generic data node configuration.
            read_fct (Optional[Callable]): The Python function called to read the data.
            write_fct (Optional[Callable]): The Python function called to write the data.
                The provided function must have at least one parameter that receives the data to be written.
            read_fct_args (Optional[List]): The list of arguments that are passed to the function
                *read_fct* to read data.
            write_fct_args (Optional[List]): The list of arguments that are passed to the function
                *write_fct* to write the data.
            scope (Optional[Scope^]): The scope of the Generic data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new Generic data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_in_memory_data_node(
        cls,
        id: str,
        default_data: Optional[Any] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new *in-memory* data node configuration.

        Parameters:
            id (str): The unique identifier of the new in_memory data node configuration.
            default_data (Optional[any]): The default data of the data nodes instantiated from
                this in_memory data node configuration.
                If provided, note that the default_data will be stored as a configuration attribute.
                So it is designed to handle small data values like parameters, and it must be Json serializable.
            scope (Optional[Scope^]): The scope of the in_memory data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new *in-memory* data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_pickle_data_node(
        cls,
        id: str,
        default_path: Optional[str] = None,
        default_data: Optional[Any] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new pickle data node configuration.

        Parameters:
            id (str): The unique identifier of the new pickle data node configuration.
            default_path (Optional[str]): The path of the pickle file.
            default_data (Optional[any]): The default data of the data nodes instantiated from
                this pickle data node configuration.
                If provided, note that the default_data will be stored as a configuration attribute.
                So it is designed to handle small data values like parameters, and it must be Json serializable.
            scope (Optional[Scope^]): The scope of the pickle data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new pickle data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_sql_table_data_node(
        cls,
        id: str,
        db_name: str,
        db_engine: str,
        table_name: str,
        db_username: Optional[str] = None,
        db_password: Optional[str] = None,
        db_host: Optional[str] = None,
        db_port: Optional[int] = None,
        db_driver: Optional[str] = None,
        sqlite_folder_path: Optional[str] = None,
        sqlite_file_extension: Optional[str] = None,
        db_extra_args: Optional[Dict[str, Any]] = None,
        exposed_type: Optional[str] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new SQL table data node configuration.

        Parameters:
            id (str): The unique identifier of the new SQL data node configuration.
            db_name (str): The database name, or the name of the SQLite database file.
            db_engine (str): The database engine. Possible values are *"sqlite"*, *"mssql"*, *"mysql"*,
                or *"postgresql"*.
            table_name (str): The name of the SQL table.
            db_username (Optional[str]): The database username. Required by the *"mssql"*, *"mysql"*, and
                *"postgresql"* engines.
            db_password (Optional[str]): The database password. Required by the *"mssql"*, *"mysql"*, and
                *"postgresql"* engines.
            db_host (Optional[str]): The database host.<br/>
                The default value is "localhost".
            db_port (Optional[int]): The database port.<br/>
                The default value is 1433.
            db_driver (Optional[str]): The database driver.
            sqlite_folder_path (Optional[str]): The path to the folder that contains SQLite file.<br/>
                The default value is the current working folder.
            sqlite_file_extension (Optional[str]): The file extension of the SQLite file.<br/>
                The default value is ".db".
            db_extra_args (Optional[dict[str, any]]): A dictionary of additional arguments to be passed
                into database connection string.
            exposed_type (Optional[str]): The exposed type of the data read from SQL table.<br/>
                The default value is "pandas".
            scope (Optional[Scope^]): The scope of the SQL data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new SQL data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_sql_data_node(
        cls,
        id: str,
        db_name: str,
        db_engine: str,
        read_query: str,
        write_query_builder: Callable,
        append_query_builder: Optional[Callable] = None,
        db_username: Optional[str] = None,
        db_password: Optional[str] = None,
        db_host: Optional[str] = None,
        db_port: Optional[int] = None,
        db_driver: Optional[str] = None,
        sqlite_folder_path: Optional[str] = None,
        sqlite_file_extension: Optional[str] = None,
        db_extra_args: Optional[Dict[str, Any]] = None,
        exposed_type: Optional[str] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new SQL data node configuration.

        Parameters:
            id (str): The unique identifier of the new SQL data node configuration.
            db_name (str): The database name, or the name of the SQLite database file.
            db_engine (str): The database engine. Possible values are *"sqlite"*, *"mssql"*, *"mysql"*,
                or *"postgresql"*.
            read_query (str): The SQL query string used to read the data from the database.
            write_query_builder (Callable): A callback function that takes the data as an input parameter
                and returns a list of SQL queries to be executed when writing data to the data node.
            append_query_builder (Optional[Callable]): A callback function that takes the data as an input parameter
                and returns a list of SQL queries to be executed when appending data to the data node.
            db_username (Optional[str]): The database username. Required by the *"mssql"*, *"mysql"*, and
                *"postgresql"* engines.
            db_password (Optional[str]): The database password. Required by the *"mssql"*, *"mysql"*, and
                *"postgresql"* engines.
            db_host (Optional[str]): The database host.<br/>
                The default value is "localhost".
            db_port (Optional[int]): The database port.<br/>
                The default value is 1433.
            db_driver (Optional[str]): The database driver.
            sqlite_folder_path (Optional[str]): The path to the folder that contains SQLite file.<br/>
                The default value is the current working folder.
            sqlite_file_extension (Optional[str]): The file extension of the SQLite file.<br/>
                The default value is ".db".
            db_extra_args (Optional[dict[str, any]]): A dictionary of additional arguments to be passed
                into database connection string.
            exposed_type (Optional[str]): The exposed type of the data read from SQL query.<br/>
                The default value is "pandas".
            scope (Optional[Scope^]): The scope of the SQL data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new SQL data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_mongo_collection_data_node(
        cls,
        id: str,
        db_name: str,
        collection_name: str,
        custom_document: Optional[Any] = None,
        db_username: Optional[str] = None,
        db_password: Optional[str] = None,
        db_host: Optional[str] = None,
        db_port: Optional[int] = None,
        db_driver: Optional[str] = None,
        db_extra_args: Optional[Dict[str, Any]] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new Mongo collection data node configuration.

        Parameters:
            id (str): The unique identifier of the new Mongo collection data node configuration.
            db_name (str): The database name.
            collection_name (str): The collection in the database to read from and to write the data to.
            custom_document (Optional[any]): The custom document class to store, encode, and decode data
                when reading and writing to a Mongo collection. The custom_document can have an optional
                *decode()* method to decode data in the Mongo collection to a custom object, and an
                optional *encode()*) method to encode the object's properties to the Mongo collection
                when writing.
            db_username (Optional[str]): The database username.
            db_password (Optional[str]): The database password.
            db_host (Optional[str]): The database host.<br/>
                The default value is "localhost".
            db_port (Optional[int]): The database port.<br/>
                The default value is 27017.
            db_driver (Optional[str]): The database driver.
            db_extra_args (Optional[dict[str, any]]): A dictionary of additional arguments to be passed
                into database connection string.
            scope (Optional[Scope^]): The scope of the Mongo collection data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new Mongo collection data node configuration.
        """  # noqa: E501

    @classmethod
    def configure_s3_object_data_node(
        cls,
        id: str,
        aws_access_key: str,
        aws_secret_access_key: str,
        aws_s3_bucket_name: str,
        aws_s3_object_key: str,
        aws_region: Optional[str] = None,
        aws_s3_object_parameters: Optional[Dict[str, Any]] = None,
        scope: Optional[Scope] = None,
        validity_period: Optional[timedelta] = None,
        **properties,
    ) -> "DataNodeConfig":
        """Configure a new S3 object data node configuration.

        Parameters:
            id (str): The unique identifier of the new S3 Object data node configuration.
            aws_access_key (str): Amazon Web Services ID for to identify account.
            aws_secret_access_key (str): Amazon Web Services access key to authenticate programmatic requests.
            aws_s3_bucket_name (str): The bucket in S3 to read from and to write the data to.
            aws_region (Optional[str]): Self-contained geographic area where Amazon Web Services (AWS)
                infrastructure is located.
            aws_s3_object_parameters (Optional[dict[str, any]]): A dictionary of additional arguments to be passed
                into AWS S3 bucket access string.
            scope (Optional[Scope^]): The scope of the S3 Object data node configuration.<br/>
                The default value is `Scope.SCENARIO`.
            validity_period (Optional[timedelta]): The duration since the last edit date for which the data node can be
                considered up-to-date. Once the validity period has passed, the data node is considered stale and
                relevant tasks will run even if they are skippable (see the Task configuration
                [page](../../../../../../userman/scenario_features/task-orchestration/scenario-config.md#from-task-configurations)
                for more details).
                If *validity_period* is set to None, the data node is always up-to-date.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new S3 object data node configuration.
        """  # noqa: E501

    @staticmethod
    def configure_task(
        id: str,
        function: Optional[Callable],
        input: Optional[Union[DataNodeConfig, List[DataNodeConfig]]] = None,
        output: Optional[Union[DataNodeConfig, List[DataNodeConfig]]] = None,
        skippable: bool = False,
        **properties,
    ) -> "TaskConfig":
        """Configure a new task configuration.

        Parameters:
            id (str): The unique identifier of this task configuration.
            function (Callable): The python function called by Taipy to run the task.
            input (Optional[Union[DataNodeConfig^, List[DataNodeConfig^]]]): The list of the
                function input data node configurations. This can be a unique data node
                configuration if there is a single input data node, or None if there are none.
            output (Optional[Union[DataNodeConfig^, List[DataNodeConfig^]]]): The list of the
                function output data node configurations. This can be a unique data node
                configuration if there is a single output data node, or None if there are none.
            skippable (bool): If True, indicates that the task can be skipped if no change has
                been made on inputs.<br/>
                The default value is False.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new task configuration.
        """

    @staticmethod
    def set_default_task_configuration(
        function: Optional[Callable],
        input: Optional[Union[DataNodeConfig, List[DataNodeConfig]]] = None,
        output: Optional[Union[DataNodeConfig, List[DataNodeConfig]]] = None,
        skippable: bool = False,
        **properties,
    ) -> "TaskConfig":
        """Set the default values for task configurations.

        This function creates the *default task configuration* object,
        where all task configuration objects will find their default
        values when needed.

        Parameters:
            function (Callable): The python function called by Taipy to run the task.
            input (Optional[Union[DataNodeConfig^, List[DataNodeConfig^]]]): The list of the
                input data node configurations. This can be a unique data node
                configuration if there is a single input data node, or None if there are none.
            output (Optional[Union[DataNodeConfig^, List[DataNodeConfig^]]]): The list of the
                output data node configurations. This can be a unique data node
                configuration if there is a single output data node, or None if there are none.
            skippable (bool): If True, indicates that the task can be skipped if no change has
                been made on inputs.<br/>
                The default value is False.
            **properties (dict[str, any]): A keyworded variable length list of additional
                arguments.

        Returns:
            The default task configuration.
        """

    @staticmethod
    def configure_job_executions(
        mode: Optional[str] = None, max_nb_of_workers: Optional[Union[int, str]] = None, **properties
    ) -> "JobConfig":
        """Configure job execution.

        Parameters:
            mode (Optional[str]): The job execution mode.
                Possible values are: *"standalone"* or *"development"*.
            max_nb_of_workers (Optional[int, str]): Parameter used only in *"standalone"* mode.
                This indicates the maximum number of jobs able to run in parallel.<br/>
                The default value is 2.<br/>
                A string can be provided to dynamically set the value using an environment
                variable. The string must follow the pattern: `ENV[&lt;env_var&gt;]` where
                `&lt;env_var&gt;` is the name of an environment variable.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            The new job execution configuration.
        """

    @staticmethod
    def configure_core(
        root_folder: Optional[str] = None,
        storage_folder: Optional[str] = None,
        taipy_storage_folder: Optional[str] = None,
        repository_type: Optional[str] = None,
        repository_properties: Optional[Dict[str, Union[str, int]]] = None,
        read_entity_retry: Optional[int] = None,
        mode: Optional[str] = None,
        version_number: Optional[str] = None,
        force: Optional[bool] = None,
        **properties,
    ) -> "CoreSection":
        """Configure the Orchestrator service.

        Parameters:
            root_folder (Optional[str]): Path of the base folder for the taipy application.
                The default value is "./taipy/"
            storage_folder (str): Folder name used to store user data. The default value is "user_data/". It is used in
                conjunction with the *root_folder* attribute. That means the storage path is
                <root_folder><storage_folder> (The default path is "./taipy/user_data/").
            taipy_storage_folder (str): Folder name used to store Taipy data. The default value is ".taipy/". It is
                used in conjunction with the *root_folder* attribute. That means the storage path is
                <root_folder><storage_folder> (The default path is "./taipy/.taipy/").
            repository_type (Optional[str]): The type of the repository to be used to store Taipy data.
                The default value is "filesystem".
            repository_properties (Optional[Dict[str, Union[str, int]]]): A dictionary of additional properties
                to be used by the repository.
            read_entity_retry (Optional[int]): Number of retries to read an entity from the repository
                before return failure. The default value is 3.
            mode (Optional[str]): Indicates the mode of the version management system.
                Possible values are *"development"* or *"experiment"*. On Enterprise edition of Taipy,
                *production* mode is also available. Please refer to the
                [Versioning management](../../../../../../userman/advanced_features/versioning/index.md)
                documentation page for more details.
            version_number (Optional[str]): The string identifier of the version.
                 In development mode, the version number is ignored.
            force (Optional[bool]): If True, Taipy will override a version even if the configuration
                has changed and run the application.
            **properties (Dict[str, Any]): A keyworded variable length list of additional arguments configure the
                behavior of the `Orchestrator^` service.

        Returns:
            The Core configuration.
        """



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import json
import os


def _get_version():
    with open(f"{os.path.dirname(os.path.abspath(__file__))}{os.sep}version.json") as version_file:
        version = json.load(version_file)
        version_string = f'{version.get("major", 0)}.{version.get("minor", 0)}.{version.get("patch", 0)}'
        if vext := version.get("ext"):
            version_string = f"{version_string}.{vext}"
    return version_string



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from abc import ABC

from .section import Section


class UniqueSection(Section, ABC):
    """An abstract class representing a subdivision of the configuration class `Config^`.

    Each UniqueSection implementation class defines a semantically consistent set of settings
    related to a particular aspect of the application. It differs from a regular `Section` in
    that it is designed to be unique, meaning only one instance can exist. Each UniqueSection
    is defined by a section name (related to the objects they configure) and a set of properties.

    Here are the various unique sections in Taipy:

    - `GlobalAppConfig^` for configuring global application settings.
    - `GuiSection` for configuring the GUI service.
    - `CoreSection^` for configuring the core package behavior.
    - `JobConfig^` for configuring the job orchestration.
    - `AuthenticationConfig^` for configuring authentication settings.

    """

    def __init__(self, **properties):
        super().__init__(self.name, **properties)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
from typing import Dict, List, Set

from .._serializer._json_serializer import _JsonSerializer


class _ComparatorResult(dict):
    ADDED_ITEMS_KEY = "added_items"
    REMOVED_ITEMS_KEY = "removed_items"
    MODIFIED_ITEMS_KEY = "modified_items"

    CONFLICTED_SECTION_KEY = "conflicted_sections"
    UNCONFLICTED_SECTION_KEY = "unconflicted_sections"

    def __init__(self, unconflicted_sections: Set[str]):
        super().__init__()

        self._unconflicted_sections = unconflicted_sections

    def _sort_by_section(self):
        if self.get(self.CONFLICTED_SECTION_KEY):
            for key in self[self.CONFLICTED_SECTION_KEY].keys():
                self[self.CONFLICTED_SECTION_KEY][key].sort(key=lambda x: x[0][0])

        if self.get(self.UNCONFLICTED_SECTION_KEY):
            for key in self[self.UNCONFLICTED_SECTION_KEY].keys():
                self[self.UNCONFLICTED_SECTION_KEY][key].sort(key=lambda x: x[0][0])

    def _check_added_items(self, config_deepdiff, new_json_config):
        if dictionary_item_added := config_deepdiff.get("dictionary_item_added"):
            for item_added in dictionary_item_added:
                section_name, config_id, attribute = self.__get_changed_entity_attribute(item_added)
                diff_sections = self.__get_section(section_name)

                if attribute:
                    value_added = new_json_config[section_name][config_id][attribute]
                elif config_id:
                    value_added = new_json_config[section_name][config_id]
                else:
                    value_added = new_json_config[section_name]

                section_name = self.__rename_global_node_name(section_name)
                self.__create_or_append_list(
                    diff_sections,
                    self.ADDED_ITEMS_KEY,
                    ((section_name, config_id, attribute), (value_added)),
                )

    def _check_removed_items(self, config_deepdiff, old_json_config):
        if dictionary_item_removed := config_deepdiff.get("dictionary_item_removed"):
            for item_removed in dictionary_item_removed:
                section_name, config_id, attribute = self.__get_changed_entity_attribute(item_removed)
                diff_sections = self.__get_section(section_name)

                if attribute:
                    value_removed = old_json_config[section_name][config_id][attribute]
                elif config_id:
                    value_removed = old_json_config[section_name][config_id]
                else:
                    value_removed = old_json_config[section_name]

                section_name = self.__rename_global_node_name(section_name)
                self.__create_or_append_list(
                    diff_sections,
                    self.REMOVED_ITEMS_KEY,
                    ((section_name, config_id, attribute), (value_removed)),
                )

    def _check_modified_items(self, config_deepdiff, old_json_config, new_json_config):
        if values_changed := config_deepdiff.get("values_changed"):
            for item_changed, value_changed in values_changed.items():
                section_name, config_id, attribute = self.__get_changed_entity_attribute(item_changed)
                diff_sections = self.__get_section(section_name)

                section_name = self.__rename_global_node_name(section_name)
                self.__create_or_append_list(
                    diff_sections,
                    self.MODIFIED_ITEMS_KEY,
                    ((section_name, config_id, attribute), (value_changed["old_value"], value_changed["new_value"])),
                )

        # Iterable item added will be considered a modified item
        if iterable_item_added := config_deepdiff.get("iterable_item_added"):
            self.__check_modified_iterable(iterable_item_added, old_json_config, new_json_config)

        # Iterable item removed will be considered a modified item
        if iterable_item_removed := config_deepdiff.get("iterable_item_removed"):
            self.__check_modified_iterable(iterable_item_removed, old_json_config, new_json_config)

    def __check_modified_iterable(self, iterable_items, old_json_config, new_json_config):
        for item in iterable_items:
            section_name, config_id, attribute = self.__get_changed_entity_attribute(item)
            diff_sections = self.__get_section(section_name)

            if attribute:
                new_value = new_json_config[section_name][config_id][attribute]
                old_value = old_json_config[section_name][config_id][attribute]
            else:
                new_value = new_json_config[section_name][config_id]
                old_value = old_json_config[section_name][config_id]

            section_name = self.__rename_global_node_name(section_name)
            modified_value = ((section_name, config_id, attribute), (old_value, new_value))

            if (
                not diff_sections.get(self.MODIFIED_ITEMS_KEY)
                or modified_value not in diff_sections[self.MODIFIED_ITEMS_KEY]
            ):
                self.__create_or_append_list(
                    diff_sections,
                    self.MODIFIED_ITEMS_KEY,
                    modified_value,
                )

    def __get_section(self, section_name: str) -> Dict[str, List]:
        if section_name in self._unconflicted_sections:
            if not self.get(self.UNCONFLICTED_SECTION_KEY):
                self[self.UNCONFLICTED_SECTION_KEY] = {}
            return self[self.UNCONFLICTED_SECTION_KEY]

        if not self.get(self.CONFLICTED_SECTION_KEY):
            self[self.CONFLICTED_SECTION_KEY] = {}
        return self[self.CONFLICTED_SECTION_KEY]

    def __create_or_append_list(self, diff_dict, key, value):
        if diff_dict.get(key):
            diff_dict[key].append(value)
        else:
            diff_dict[key] = [value]

    def __get_changed_entity_attribute(self, attribute_bracket_notation):
        """Split the section name, the config id (if exists), and the attribute name (if exists)
        from JSON bracket notation.
        """
        try:
            section_name, config_id, attribute = re.findall(r"\[\'(.*?)\'\]", attribute_bracket_notation)
        except ValueError:
            try:
                section_name, config_id = re.findall(r"\[\'(.*?)\'\]", attribute_bracket_notation)
                attribute = None
            except ValueError:
                section_name = re.findall(r"\[\'(.*?)\'\]", attribute_bracket_notation)[0]
                config_id = None
                attribute = None

        return section_name, config_id, attribute

    def __rename_global_node_name(self, node_name):
        if node_name == _JsonSerializer._GLOBAL_NODE_NAME:
            return "Global Configuration"
        return node_name



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ..._config import _Config
from ..issue_collector import IssueCollector
from ._config_checker import _ConfigChecker


class _AuthConfigChecker(_ConfigChecker):
    def __init__(self, config: _Config, collector: IssueCollector):
        super().__init__(config, collector)

    def _check(self) -> IssueCollector:
        auth_config = self._config._auth_config  # type: ignore
        self._check_predefined_protocol(auth_config)
        return self._collector

    def _check_predefined_protocol(self, auth_config):
        if auth_config.protocol == auth_config._PROTOCOL_LDAP:
            self.__check_ldap(auth_config)
        if auth_config.protocol == auth_config._PROTOCOL_TAIPY:
            self.__check_taipy(auth_config)

    def __check_taipy(self, auth_config):
        if auth_config._TAIPY_ROLES not in auth_config.properties:
            self._error(
                "properties",
                auth_config._LDAP_SERVER,
                f"`{auth_config._LDAP_SERVER}` property must be populated when {auth_config._PROTOCOL_LDAP} is used.",
            )
        if auth_config._TAIPY_PWD not in auth_config.properties:
            self._warning(
                "properties",
                auth_config._TAIPY_PWD,
                f"`In order to protect authentication with passwords using {auth_config._PROTOCOL_TAIPY} protocol,"
                f" {auth_config._TAIPY_PWD}` property can be populated.",
            )

    def __check_ldap(self, auth_config):
        if auth_config._LDAP_SERVER not in auth_config.properties:
            self._error(
                "properties",
                auth_config._LDAP_SERVER,
                f"`{auth_config._LDAP_SERVER}` attribute must be populated when {auth_config._PROTOCOL_LDAP} is used.",
            )
        if auth_config._LDAP_BASE_DN not in auth_config.properties:
            self._error(
                "properties",
                auth_config._LDAP_BASE_DN,
                f"`{auth_config._LDAP_BASE_DN}` field must be populated when {auth_config._PROTOCOL_LDAP} is used.",
            )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from abc import abstractmethod
from typing import Any, Dict, Optional

from .common._config_blocker import _ConfigBlocker
from .common._template_handler import _TemplateHandler as _tpl
from .common._validate_id import _validate_id


class Section:
    """An abstract class representing a subdivision of the configuration class `Config^`.

    The role of the subclasses of this class is to define semantically consistent sets of settings
    related to a particular aspect of the application.

    Here are the various sections in Taipy:

    - `DataNodeConfig^` for configuring data nodes.
    - `TaskConfig^` for configuring tasks.
    - `ScenarioConfig^` for configuring scenarios.
    - `MigrationConfig^` for configuring data migration within the Taipy version management.

    Each Section implementation is defined by a section name (related to the objects they
    configure), a unique identifier, and a set of properties.

    Some of the Section implementations are designed to be unique, meaning only one instance
    can exist. They are subclasses of the `UniqueSection^` abstract class such as:

    - `GlobalAppConfig^` for configuring global application settings.
    - `_GuiConfig` for configuring the GUI service.
    - `CoreSection^` for configuring the core package behavior.
    - `JobConfig^` for configuring the job orchestration.
    - `AuthenticationConfig^` for configuring authentication settings.

    """

    _DEFAULT_KEY = "default"
    _ID_KEY = "id"

    id: str
    """A valid python identifier that uniquely identifies the section."""

    def __init__(self, id, **properties):
        self.id = _validate_id(id)
        self._properties = properties or {}

    @abstractmethod
    def __copy__(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self):
        """The name of the section.

        This property is used to identify the section in the configuration. It is used as a key in the
        dictionary of sections in the `Config^` class.
        Note also that the name of the section is exposed as a `Config^` property.
        """
        raise NotImplementedError

    @abstractmethod
    def _clean(self):
        raise NotImplementedError

    @abstractmethod
    def _to_dict(self):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _from_dict(cls, config_as_dict: Dict[str, Any], id, config):
        raise NotImplementedError

    @abstractmethod
    def _update(self, config_as_dict, default_section=None):
        raise NotImplementedError

    def __getattr__(self, item: str) -> Optional[Any]:
        return self._replace_templates(self._properties.get(item, None))

    @property
    def properties(self):
        """A dictionary of additional properties."""
        return {k: _tpl._replace_templates(v) for k, v in self._properties.items()}

    @properties.setter  # type: ignore
    @_ConfigBlocker._check()
    def properties(self, val):
        self._properties = val

    def _replace_templates(self, value):
        return _tpl._replace_templates(value)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import toml  # type: ignore

from .._config import _Config
from ..exceptions.exceptions import LoadingError
from ._base_serializer import _BaseSerializer


class _TomlSerializer(_BaseSerializer):
    """Convert configuration from TOML representation to Python Dict and reciprocally."""

    @classmethod
    def _write(cls, configuration: _Config, filename: str):
        with open(filename, "w") as fd:
            toml.dump(cls._str(configuration), fd)

    @classmethod
    def _read(cls, filename: str) -> _Config:
        try:
            config_as_dict = cls._pythonify(dict(toml.load(filename)))
            return cls._from_dict(config_as_dict)
        except toml.TomlDecodeError as e:
            error_msg = f"Can not load configuration {e}"
            raise LoadingError(error_msg) from None

    @classmethod
    def _serialize(cls, configuration: _Config) -> str:
        return toml.dumps(cls._str(configuration))

    @classmethod
    def _deserialize(cls, config_as_string: str) -> _Config:
        return cls._from_dict(cls._pythonify(dict(toml.loads(config_as_string))))



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import sys

from ._base_cli._abstract_cli import _AbstractCLI
from ._base_cli._taipy_parser import _TaipyParser


class _HelpCLI(_AbstractCLI):
    _COMMAND_NAME = "help"

    @classmethod
    def create_parser(cls):
        create_parser = _TaipyParser._add_subparser(
            cls._COMMAND_NAME,
            help="Show the Taipy help message.",
            add_help=False,
        )
        create_parser.add_argument(
            "command", nargs="?", type=str, const="", default="", help="Show the help message of the command."
        )

    @classmethod
    def handle_command(cls):
        args = cls._parse_arguments()
        if not args:
            return

        if args.command:
            if args.command in _TaipyParser._sub_taipyparsers.keys():
                _TaipyParser._sub_taipyparsers.get(args.command).print_help()
            else:
                cls._logger.error(f"{args.command} is not a valid command.")
        else:
            _TaipyParser._parser.print_help()

        sys.exit(0)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

""" The `taipy.common.config` package provides features to configure a Taipy application.

Its main class is the `Config^` singleton. It exposes various static methods
and attributes to configure the Taipy application and retrieve the configuration values.

!!! example "Standard usage"

    ```python
    from taipy.common.config import Config
    from taipy.common.config import Frequency
    from taipy.common.config import Scope

    data_node_cfg = Config.configure_data_node("my_data_node", scope=Scope.SCENARIO)
    Config.configure_scenario("my_scenario", additional_data_node_configs=[data_node_cfg], frequency=Frequency.DAILY)
    Config.configure_core(repository_type="filesystem", storage_folder="my/storage/folder")
    Config.configure_authentication(protocol="taipy",roles={"user1": ["role1", "TAIPY_READER"]})

    print(Config.data_nodes["my_data_node"].scope)  # Output: SCENARIO
    print(len(Config.scenarios["my_scenario"].data_nodes))  # Output: 1
    ```

    In this example, the static methods of the `Config^` singleton are used to configure
    a Taipy application. The application has one data node configuration and one scenario
    configuration.
    We also configure the application to use a filesystem repository and set up authentication.

!!! note "`Frequency^` and `Scope^` for scenario and data nodes configurations"

    Besides the `Config^` class which is the main entrypoint, the `taipy.common.config` package exposes
    the `Frequency^` and `Scope^` enums that are frequently used to configure data nodes and
    scenarios.

    The three objects are exposed in the `taipy^` package directly for convenience.

"""

from typing import List

from ._init import Config
from .common.frequency import Frequency
from .common.scope import Scope
from .global_app.global_app_config import GlobalAppConfig
from .section import Section
from .unique_section import UniqueSection


def _config_doc(func):
    def func_with_doc(section, attr_name, default, configuration_methods, add_to_unconflicted_sections=False):
        import os

        if os.environ.get("GENERATING_TAIPY_DOC", None) and os.environ["GENERATING_TAIPY_DOC"] == "true":
            with (open("config_doc.txt", "a") as f):
                from inspect import signature

                # Add the documentation for configure methods
                for exposed_configuration_method, configuration_method in configuration_methods:
                    annotation = "    @staticmethod\n"
                    sign = "    def " + exposed_configuration_method + str(signature(configuration_method)) + ":\n"
                    doc = '        """' + configuration_method.__doc__ + '"""\n'
                    content = "        pass\n\n"
                    f.write(annotation + sign + doc + content)

                # Add the documentation for the attribute
                annotation = '    @property\n'
                sign = f"    def {attr_name} (self) -> {section.__name__}:\n"
                if issubclass(section, UniqueSection):
                    doc = f'        """The configured {section.__name__} section."""\n'
                elif issubclass(section, Section):
                    doc = f'        """The configured {section.__name__} sections ."""\n'
                else:
                    print(f" ERROR - Invalid section class: {section.__name__}")  # noqa: T201
                    return
                content = "        pass\n\n"
                f.write(annotation + sign + doc + content)
        return func(section, attr_name, default, configuration_methods, add_to_unconflicted_sections)

    return func_with_doc


@_config_doc
def _inject_section(
    section_clazz,
    attribute_name: str,
    default: Section,
    configuration_methods: List[tuple],
    add_to_unconflicted_sections: bool = False,
):
    Config._register_default(default)

    if issubclass(section_clazz, UniqueSection):
        setattr(Config, attribute_name, Config.unique_sections[section_clazz.name])
    elif issubclass(section_clazz, Section):
        setattr(Config, attribute_name, Config.sections[section_clazz.name])
    else:
        raise TypeError

    if add_to_unconflicted_sections:
        Config._comparator._add_unconflicted_section(section_clazz.name)  # type: ignore[attr-defined]

    for exposed_configuration_method, configuration_method in configuration_methods:
        setattr(Config, exposed_configuration_method, configuration_method)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""The setup script."""
import json
import os
from setuptools import find_namespace_packages, find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

version_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "version.json")

with open(version_path) as version_file:
    version = json.load(version_file)
    version_string = f'{version.get("major", 0)}.{version.get("minor", 0)}.{version.get("patch", 0)}'
    if vext := version.get("ext"):
        version_string = f"{version_string}.{vext}"

requirements = ["toml>=0.10,<0.11", "deepdiff>=6.2,<6.3"]

test_requirements = ["pytest>=6.0"]

setup(
    version=version_string,
    install_requires=requirements,
    packages=find_namespace_packages(where=".")+ find_packages(
        include=[
            "taipy",
            "taipy.common",
            "taipy.common.*",
            "taipy.common.config",
            "taipy.common.config.*",
            "taipy.common.logger",
            "taipy.common.logger.*",
            "taipy.common._cli",
            "taipy.common._cli.*"
        ]
    ),
    include_package_data=True,
    data_files=[('version', ['version.json'])],
    tests_require=test_requirements,
)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import abc
from typing import Any, List, Optional, Set

from ..._config import _Config
from ..issue_collector import IssueCollector


class _ConfigChecker:
    _PREDEFINED_PROPERTIES_KEYS = ["_entity_owner"]

    def __init__(self, config: _Config, collector):
        self._collector = collector
        self._config = config

    @abc.abstractmethod
    def _check(self) -> IssueCollector:
        raise NotImplementedError

    def _error(self, field: str, value: Any, message: str):
        self._collector._add_error(field, value, message, self.__class__.__name__)

    def _warning(self, field: str, value: Any, message: str):
        self._collector._add_warning(field, value, message, self.__class__.__name__)

    def _info(self, field: str, value: Any, message: str):
        self._collector._add_info(field, value, message, self.__class__.__name__)

    def _check_children(
        self,
        parent_config_class,
        config_id: str,
        config_key: str,
        config_value,
        child_config_class,
        can_be_empty: Optional[bool] = False,
    ):
        if not config_value and not can_be_empty:
            self._warning(
                config_key,
                config_value,
                f"{config_key} field of {parent_config_class.__name__} `{config_id}` is empty.",
            )
        elif not (
            (isinstance(config_value, (List, Set))) and all(isinstance(x, child_config_class) for x in config_value)
        ):
            self._error(
                config_key,
                config_value,
                f"{config_key} field of {parent_config_class.__name__} `{config_id}` must be populated with a list "
                f"of {child_config_class.__name__} objects.",
            )

    def _check_existing_config_id(self, config):
        if not config.id:
            self._error(
                "config_id",
                config.id,
                f"config_id of {config.__class__.__name__} `{config.id}` is empty.",
            )

    def _check_if_entity_property_key_used_is_predefined(self, config):
        for key, value in config._properties.items():
            if key in self._PREDEFINED_PROPERTIES_KEYS:
                self._error(
                    key,
                    value,
                    f"Properties of {config.__class__.__name__} `{config.id}` cannot have `{key}` as its property.",
                )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import subprocess
import sys

from ._base_cli._abstract_cli import _AbstractCLI
from ._base_cli._taipy_parser import _TaipyParser


class _RunCLI(_AbstractCLI):
    _COMMAND_NAME = "run"

    @classmethod
    def create_parser(cls):
        run_parser = _TaipyParser._add_subparser(cls._COMMAND_NAME, help="Run a Taipy application.")
        run_parser.add_argument(
            "application_main_file",
        )

        sub_run_parser = run_parser.add_subparsers(title="subcommands")
        sub_run_parser.add_parser(
            "external-args",
            help="""
                Arguments defined after this keyword will be considered as external arguments
                to be passed to the application
            """,
        )

    @classmethod
    def handle_command(cls):
        args, _ = _TaipyParser._parser.parse_known_args()
        if getattr(args, "which", None) != "run":
            return

        # First 2 args are always (1) Python executable, (2) run
        # Unknown args are passed when running the application but will be ignored
        all_args = sys.argv[2:]

        external_args = []
        try:
            external_args_index = all_args.index("external-args")
        except ValueError:
            pass
        else:
            external_args.extend(all_args[external_args_index + 1 :])
            all_args = all_args[:external_args_index]

        taipy_args = [f"--taipy-{arg[2:]}" if arg.startswith("--") else arg for arg in all_args]

        try:
            subprocess.run(
                [sys.executable, args.application_main_file, *(external_args + taipy_args)],
                stdout=sys.stdout,
                stderr=sys.stdout,
            )
        except KeyboardInterrupt:
            pass

        sys.exit(0)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from copy import copy
from typing import Dict

from .global_app.global_app_config import GlobalAppConfig
from .section import Section
from .unique_section import UniqueSection


class _Config:
    DEFAULT_KEY = "default"

    def __init__(self) -> None:
        self._sections: Dict[str, Dict[str, Section]] = {}
        self._unique_sections: Dict[str, UniqueSection] = {}
        self._global_config: GlobalAppConfig = GlobalAppConfig()

    def _clean(self) -> None:
        self._global_config._clean()
        for unique_section in self._unique_sections.values():
            unique_section._clean()
        for sections in self._sections.values():
            for section in sections.values():
                section._clean()

    @classmethod
    def _default_config(cls):
        config = _Config()
        config._global_config = GlobalAppConfig.default_config()
        return config

    def _update(self, other_config) -> None:
        self._global_config._update(other_config._global_config._to_dict())
        if other_config._unique_sections:
            for section_name, other_section in other_config._unique_sections.items():
                if section := self._unique_sections.get(section_name, None):
                    section._update(other_section._to_dict())
                else:
                    self._unique_sections[section_name] = copy(other_config._unique_sections[section_name])
        if other_config._sections:
            for section_name, other_non_unique_sections in other_config._sections.items():
                if non_unique_sections := self._sections.get(section_name, None):
                    self.__update_sections(non_unique_sections, other_non_unique_sections)
                else:
                    self._sections[section_name] = {}
                    self.__add_sections(self._sections[section_name], other_non_unique_sections)

    def __add_sections(self, entity_config, other_entity_configs) -> None:
        for cfg_id, sub_config in other_entity_configs.items():
            entity_config[cfg_id] = copy(sub_config)
            self.__point_nested_section_to_self(sub_config)

    def __update_sections(self, entity_config, other_entity_configs) -> None:
        if self.DEFAULT_KEY in other_entity_configs:
            if self.DEFAULT_KEY in entity_config:
                entity_config[self.DEFAULT_KEY]._update(other_entity_configs[self.DEFAULT_KEY]._to_dict())
            else:
                entity_config[self.DEFAULT_KEY] = other_entity_configs[self.DEFAULT_KEY]
        for cfg_id, sub_config in other_entity_configs.items():
            if cfg_id != self.DEFAULT_KEY:
                if cfg_id not in entity_config:
                    entity_config[cfg_id] = copy(sub_config)
                entity_config[cfg_id]._update(sub_config._to_dict(), entity_config.get(self.DEFAULT_KEY))
            self.__point_nested_section_to_self(sub_config)

    def __point_nested_section_to_self(self, section) -> None:
        """Loop through attributes of a Section to find if any attribute has a list of Section as value.
        If there is, update each nested Section by the corresponding instance in self.

        Args:
            section (Section): The Section to search for nested sections.
        """
        for _, attr_value in vars(section).items():
            # ! This will fail if an attribute is a dictionary, or nested list of Sections.
            if not isinstance(attr_value, list):
                continue

            for index, item in enumerate(attr_value):
                if not isinstance(item, Section):
                    continue

                if sub_item := self._sections.get(item.name, {}).get(item.id, None):
                    attr_value[index] = sub_item



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ..common._repr_enum import _ReprEnum


class _OrderedEnum(_ReprEnum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class Scope(_OrderedEnum):
    """Scope of a `DataNodeConfig^` or a `DataNode^`.

    This enumeration can have the following values:

    - `GLOBAL`: Global scope, the data node is shared by all the scenarios.
    - `CYCLE`: Cycle scope, the data node is shared by all the scenarios of the same cycle.
    - `SCENARIO` (Default value): Scenario scope, the data node is unique to a scenario.

    Each data node config has a scope. It is an attribute propagated to the `DataNode^`
    when instantiated from a `DataNodeConfig^`. The scope is used to determine the
    _visibility_ of the data node, and which scenarios can access it.

    In other words :

    - There can be only one data node instantiated from a `DataNodeConfig^` with a `GLOBAL`
        scope. All the scenarios share the unique data node. When a new scenario is created,
        the data node is also created if and only if it does not exist yet.
    - Only one data node instantiated from a `DataNodeConfig^` with a `CYCLE` scope is
        created for each cycle. All the scenarios of the same cycle share the same data node.
        When a new scenario is created within a cycle, Taipy instantiates a new data node if
        and only if there is no data node for the cycle yet.
    - A data node that has the scope set to `SCENARIO` belongs to a unique scenario and cannot
        be used by others. When creating a new scenario, data nodes with a `SCENARIO` scope
        are systematically created along with the new scenario.

    !!! example

        Let's consider a simple example where a company wants to predict its sales for the next
        month. The company has a trained model that predicts the sales based on the current month
        and the historical sales. Based on the sales forecasts the company wants to plan its
        production orders. The company wants to simulate two scenarios every month: one with
        low capacity and one with high capacity.

        We can create the `DataNodeConfig^`s with the following scopes:

        - One data node for the historical sales with a `GLOBAL` scope.
        - Three data nodes with a `CYCLE` scope, for the trained model, the current month,
            and the sales predictions.
        - Two data nodes with a `SCENARIO` scope, for the capacity and the production orders.

        The code snippet below shows how to configure the data nodes with the different scopes:

        ```python
        from taipy import Config, Scope

        hist_cfg = Config.configure_csv_data_node("sales_history", scope=Scope.GLOBAL)
        model_cfg = Config.configure_data_node("trained_model", scope=Scope.CYCLE)
        month_cfg = Config.configure_data_node("current_month", scope=Scope.CYCLE)
        predictions_cfg = Config.configure_data_node("sales_predictions", scope=Scope.CYCLE)
        capacity_cfg = Config.configure_data_node("capacity", scope=Scope.SCENARIO)
        orders_cfg = Config.configure_sql_data_node("production_orders",
                                                    scope=Scope.SCENARIO,
                                                    db_name="taipy",
                                                    db_engine="sqlite",
                                                    table_name="sales")
        ```
    """

    GLOBAL = 3
    CYCLE = 2
    SCENARIO = 1



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import functools

from ...logger._taipy_logger import _TaipyLogger
from ..exceptions.exceptions import ConfigurationUpdateBlocked


class _ConfigBlocker:
    """Configuration blocker singleton."""

    __logger = _TaipyLogger._get_logger()
    __block_config_update = False

    @classmethod
    def _block(cls):
        if not cls.__block_config_update:
            cls.__logger.debug("Blocking configuration update.")
            cls.__block_config_update = True

    @classmethod
    def _unblock(cls):
        if cls.__block_config_update:
            cls.__logger.debug("Unblocking configuration update.")
            cls.__block_config_update = False

    @classmethod
    def _check(cls):
        def inner(f):
            @functools.wraps(f)
            def _check_if_is_blocking(*args, **kwargs):
                if cls.__block_config_update:
                    error_message = (
                        "The Orchestrator service should be stopped by running orchestrator.stop() before"
                        " modifying the Configuration."
                    )
                    cls.__logger.error(f"ConfigurationUpdateBlocked: {error_message}")
                    raise ConfigurationUpdateBlocked(error_message)

                return f(*args, **kwargs)

            return _check_if_is_blocking

        return inner



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import json
from copy import copy
from typing import Optional, Set, Union

from deepdiff import DeepDiff

from ...logger._taipy_logger import _TaipyLogger
from .._config import _Config
from .._serializer._json_serializer import _JsonSerializer
from ._comparator_result import _ComparatorResult


class _ConfigComparator:
    def __init__(self) -> None:
        self._unconflicted_sections: Set[str] = set()
        self.__logger = _TaipyLogger._get_logger()

    def _add_unconflicted_section(self, section_name: Union[str, Set[str]]) -> None:
        if isinstance(section_name, str):
            section_name = {section_name}

        self._unconflicted_sections.update(section_name)

    def _find_conflict_config(
        self,
        old_config: _Config,
        new_config: _Config,
        old_version_number: Optional[str] = None,
        new_version_number: Optional[str] = None,
    ):
        """Compare between 2 _Config object to check for compatibility.

        Args:
            old_config (_Config): The old _Config.
            new_config (_Config): The new _Config.
            old_version_number (str, optional): The old version number for logging. Defaults to None.
            new_version_number (str, optional): The new version number for logging. Defaults to None.

        Returns:
            _ComparatorResult: Return a _ComparatorResult dictionary with the following format:
        ```python
        {
            "added_items": [
                ((section_name_1, config_id_1, attribute_1), added_object_1),
                ((section_name_2, config_id_2, attribute_2), added_object_2),
            ],
            "removed_items": [
                ((section_name_1, config_id_1, attribute_1), removed_object_1),
                ((section_name_2, config_id_2, attribute_2), removed_object_2),
            ],
            "modified_items": [
                ((section_name_1, config_id_1, attribute_1), (old_value_1, new_value_1)),
                ((section_name_2, config_id_2, attribute_2), (old_value_2, new_value_2)),
            ],
        }
        ```
        """
        comparator_result = self.__get_config_diff(old_config, new_config)
        self.__log_find_conflict_message(comparator_result, old_version_number, new_version_number)
        return comparator_result

    def _compare(
        self,
        config_1: _Config,
        config_2: _Config,
        version_number_1: str,
        version_number_2: str,
    ):
        """Compare between 2 _Config object to check for compatibility.

        Args:
            config_1 (_Config): The old _Config.
            config_2 (_Config): The new _Config.
            version_number_1 (str): The old version number for logging.
            version_number_2 (str): The new version number for logging.
        """
        comparator_result = self.__get_config_diff(config_1, config_2)
        self.__log_comparison_message(comparator_result, version_number_1, version_number_2)

        return comparator_result

    def __get_config_diff(self, config_1, config_2):
        json_config_1 = json.loads(_JsonSerializer._serialize(config_1))
        json_config_2 = json.loads(_JsonSerializer._serialize(config_2))

        config_deepdiff = DeepDiff(json_config_1, json_config_2, ignore_order=True)

        comparator_result = _ComparatorResult(copy(self._unconflicted_sections))

        comparator_result._check_added_items(config_deepdiff, json_config_2)
        comparator_result._check_removed_items(config_deepdiff, json_config_1)
        comparator_result._check_modified_items(config_deepdiff, json_config_1, json_config_2)
        comparator_result._sort_by_section()

        return comparator_result

    def __log_comparison_message(
        self,
        comparator_result: _ComparatorResult,
        version_number_1: str,
        version_number_2: str,
    ):
        config_str_1 = f"version {version_number_1} Configuration"
        config_str_2 = f"version {version_number_2} Configuration"

        diff_messages = []
        for _, sections in comparator_result.items():
            diff_messages = self.__get_messages(sections)

        if diff_messages:
            self.__logger.info(
                f"Differences between {config_str_1} and {config_str_2}:\n\t" + "\n\t".join(diff_messages)
            )
        else:
            self.__logger.info(f"There is no difference between {config_str_1} and {config_str_2}.")

    def __log_find_conflict_message(
        self,
        comparator_result: _ComparatorResult,
        old_version_number: Optional[str] = None,
        new_version_number: Optional[str] = None,
    ):
        old_config_str = (
            f"configuration for version {old_version_number}" if old_version_number else "current configuration"
        )
        new_config_str = (
            f"configuration for version {new_version_number}" if new_version_number else "current configuration"
        )

        if unconflicted_sections := comparator_result.get(_ComparatorResult.UNCONFLICTED_SECTION_KEY):
            unconflicted_messages = self.__get_messages(unconflicted_sections)
            self.__logger.info(
                f"There are non-conflicting changes between the {old_config_str}"
                f" and the {new_config_str}:\n\t" + "\n\t".join(unconflicted_messages)
            )

        if conflicted_sections := comparator_result.get(_ComparatorResult.CONFLICTED_SECTION_KEY):
            conflicted_messages = self.__get_messages(conflicted_sections)
            self.__logger.error(
                f"The {old_config_str} conflicts with the {new_config_str}:\n\t" + "\n\t".join(conflicted_messages)
            )

    def __get_messages(self, diff_sections):
        dq = '"'
        messages = []

        if added_items := diff_sections.get(_ComparatorResult.ADDED_ITEMS_KEY):
            for diff in added_items:
                ((section_name, config_id, attribute), added_object) = diff
                messages.append(
                    f"{section_name} {dq}{config_id}{dq} "
                    f"{f'has attribute {dq}{attribute}{dq}' if attribute else 'was'} added: {added_object}"
                )

        if removed_items := diff_sections.get(_ComparatorResult.REMOVED_ITEMS_KEY):
            for diff in removed_items:
                ((section_name, config_id, attribute), removed_object) = diff
                messages.append(
                    f"{section_name} {dq}{config_id}{dq} "
                    f"{f'has attribute {dq}{attribute}{dq}' if attribute else 'was'} removed"
                )

        if modified_items := diff_sections.get(_ComparatorResult.MODIFIED_ITEMS_KEY):
            for diff in modified_items:
                ((section_name, config_id, attribute), (old_value, new_value)) = diff
                messages.append(
                    f"{section_name} {dq}{config_id}{dq} "
                    f"{f'has attribute {dq}{attribute}{dq}' if attribute else 'was'} modified: "
                    f"{old_value} -> {new_value}"
                )

        return messages



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import argparse
from typing import Dict


class _TaipyParser:
    """Argument parser for Taipy application."""

    # The conflict_handler is set to "resolve" to override conflict arguments
    _subparser_action = None
    _parser = argparse.ArgumentParser(conflict_handler="resolve")

    _sub_taipyparsers: Dict[str, argparse.ArgumentParser] = {}
    _arg_groups: Dict[str, argparse._ArgumentGroup] = {}

    @classmethod
    def _add_subparser(cls, name: str, **kwargs) -> argparse.ArgumentParser:
        """Create a new subparser and return a argparse handler."""
        if subparser := cls._sub_taipyparsers.get(name):
            return subparser

        if not cls._subparser_action:
            cls._subparser_action = cls._parser.add_subparsers()

        subparser = cls._subparser_action.add_parser(
            name=name,
            conflict_handler="resolve",
            **kwargs,
        )
        cls._sub_taipyparsers[name] = subparser
        subparser.set_defaults(which=name)
        return subparser

    @classmethod
    def _add_groupparser(cls, title: str, description: str = "") -> argparse._ArgumentGroup:
        """Create a new group for arguments and return a argparser handler."""
        if groupparser := cls._arg_groups.get(title):
            return groupparser

        groupparser = cls._parser.add_argument_group(title=title, description=description)
        cls._arg_groups[title] = groupparser
        return groupparser

    @classmethod
    def _remove_argument(cls, arg: str):
        """Remove an argument from the parser. Note that the `arg` must be without --.

        Source: https://stackoverflow.com/questions/32807319/disable-remove-argument-in-argparse
        """
        for action in cls._parser._actions:
            opts = action.option_strings
            if (opts and opts[0] == arg) or action.dest == arg:
                cls._parser._remove_action(action)
                break

        for argument_group in cls._parser._action_groups:
            for group_action in argument_group._group_actions:
                opts = group_action.option_strings
                if (opts and opts[0] == arg) or group_action.dest == arg:
                    argument_group._group_actions.remove(group_action)
                    return



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
from typing import Dict

from ..logger._taipy_logger import _TaipyLogger
from ._config import _Config
from ._config_comparator._config_comparator import _ConfigComparator
from ._serializer._json_serializer import _JsonSerializer
from ._serializer._toml_serializer import _TomlSerializer
from .checker._checker import _Checker
from .checker.issue_collector import IssueCollector
from .common._classproperty import _Classproperty
from .common._config_blocker import _ConfigBlocker
from .global_app.global_app_config import GlobalAppConfig
from .section import Section
from .unique_section import UniqueSection


class Config:
    """Singleton class that manages the configuration of a Taipy application.

    The `Config` singleton is the main class to use for configuring a Taipy application.
    In particular, this class provides:

    1. Various methods to configure the application's behavior

        The Config class provides various methods to configure the application. Each method adds
        a specific section to the configuration and returns it.

        ??? example "Most frequently used configuration methods"

            ```python
            from taipy import Config

            def by_two(x: int):
                return x * 2

            input_cfg = Config.configure_data_node("my_input")
            result_cfg = Config.configure_data_node("my_result")
            task_cfg = Config.configure_task("my_double", function=by_two, input=input_cfg, output=result_cfg)
            scenario_cfg = Config.configure_scenario("my_scenario", task_configs=[task_cfg])

            Config.load("config.toml") # Load a configuration file
            ```

        ??? example "Advanced use case"

            The configuration can be done in three ways: Python code, configuration files, or
            environment variables. All configuration manners are ultimately merged (overriding the previous
            way) to create a final applied configuration. Please refer to the
            [advanced configuration](../../../../../../userman/advanced_features/configuration/advanced-config.md)
            section from the user manual for more details.

    2. Attributes and methods to retrieve the configuration values.

        Once the configuration is done, you can retrieve the configuration values using the exposed
        attributes.

        ??? Example "Retrieve configuration values"

            ```python
            from taipy import Config

            global_cfg = Config.global_config  # Retrieve the global application configuration
            data_node_cfgs = Config.data_nodes  # Retrieve all data node configurations
            scenario_cfgs = Config.scenarios  # Retrieve all scenario configurations
            ```

    3. A few methods to manage the configuration:

        The Config class also provides a few methods to manage the configuration.

        ??? example "Manage the configuration"

            - *Check the configuration for issues*: Use the `Config.check()^` method to check the
                configuration. It returns an `IssueCollector^` containing all the
                 `Issue^`s found. The issues are logged to the console for debugging.
            - *Block the configuration update*: Use the `Config.block_update()^` method to forbid
                any update on the configuration. This can be useful when you want to ensure that
                the configuration is not modified at run time. Note that running the `Orchestrator^`
                service` automatically blocks the configuration update.
            - *Unblock the configuration update*: Use the `Config.unblock_update()^` method to allow
                again the update on the configuration.
            - *Backup the configuration*: Use the `Config.backup()^` method to back up as a TOML
                file the applied configuration. The applied configuration backed up is the result
                of the compilation of the three possible configuration methods that overrides each
                others.
            - *Restore the configuration*: Use the `Config.restore()^` method to restore a TOML
                configuration file and replace the current applied configuration.
            - *Export the configuration*: Use the `Config.export()^` method to export as a TOML file
                the Python code configuration.
            - *Load the configuration*: Use the `Config.load()^` method to load a TOML configuration
                file and replace the current Python configuration.
            - *Override the configuration*: Use the `Config.override()^` method to load a TOML
                configuration file and override the current Python configuration.
    """

    _ENVIRONMENT_VARIABLE_NAME_WITH_CONFIG_PATH = "TAIPY_CONFIG_PATH"
    __logger = _TaipyLogger._get_logger()
    _default_config = _Config._default_config()
    _python_config = _Config()
    _file_config = _Config()
    _env_file_config = _Config()
    _applied_config = _Config()
    _collector = IssueCollector()
    _serializer = _TomlSerializer()
    __json_serializer = _JsonSerializer()
    _comparator: _ConfigComparator = _ConfigComparator()

    @_Classproperty
    def unique_sections(cls) -> Dict[str, UniqueSection]:
        """A dictionary containing all unique sections."""
        return cls._applied_config._unique_sections

    @_Classproperty
    def sections(cls) -> Dict[str, Dict[str, Section]]:
        """A dictionary containing all non-unique sections."""
        return cls._applied_config._sections

    @_Classproperty
    def global_config(cls) -> GlobalAppConfig:
        """configuration values related to the global application as a `GlobalAppConfig^`."""
        return cls._applied_config._global_config

    @classmethod
    @_ConfigBlocker._check()
    def load(cls, filename: str) -> None:
        """Load a configuration file.

        The current Python configuration is replaced and the Config compilation is triggered.

        Parameters:
            filename (Union[str, Path]): The path of the toml configuration file to load.
        """
        cls.__logger.info(f"Loading configuration. Filename: '{filename}'")
        cls._python_config = cls._serializer._read(filename)
        cls._compile_configs()
        cls.__logger.info(f"Configuration '{filename}' successfully loaded.")

    @classmethod
    def export(cls, filename: str) -> None:
        """Export a configuration.

        The export is done in a toml file. The exported configuration is taken
        from the Python code configuration.

        Note:
            If *filename* already exists, it is overwritten.

        Parameters:
            filename (Union[str, Path]): The path of the file to export.
        """
        cls._serializer._write(cls._python_config, filename)

    @classmethod
    def backup(cls, filename: str) -> None:
        """Backup a configuration.

        The backup is done in a toml file.

        The backed up configuration is a compilation from the three possible methods to configure
        the application: the Python code configuration, the file configuration and the environment
        configuration.

        Note:
            If *filename* already exists, it is overwritten.

        Parameters:
            filename (Union[str, Path]): The path of the file to export.
        """
        cls._serializer._write(cls._applied_config, filename)

    @classmethod
    @_ConfigBlocker._check()
    def restore(cls, filename: str) -> None:
        """Restore a configuration file and replace the current applied configuration.

        Parameters:
            filename (Union[str, Path]): The path of the toml configuration file to load.
        """
        cls.__logger.info(f"Restoring configuration. Filename: '{filename}'")
        cls._applied_config = cls._serializer._read(filename)
        cls.__logger.info(f"Configuration '{filename}' successfully restored.")

    @classmethod
    @_ConfigBlocker._check()
    def override(cls, filename: str) -> None:
        """Load a configuration from a file and overrides the current config.

        Parameters:
            filename (Union[str, Path]): The path of the toml configuration file to load.
        """
        cls.__logger.info(f"Loading configuration. Filename: '{filename}'")
        cls._file_config = cls._serializer._read(filename)
        cls.__logger.info("Overriding configuration.'")
        cls._compile_configs()
        cls.__logger.info(f"Configuration '{filename}' successfully loaded.")

    @classmethod
    def block_update(cls) -> None:
        """Block update on the configuration signgleton."""
        _ConfigBlocker._block()

    @classmethod
    def unblock_update(cls) -> None:
        """Unblock update on the configuration signgleton."""
        _ConfigBlocker._unblock()

    @classmethod
    @_ConfigBlocker._check()
    def configure_global_app(cls, **properties) -> GlobalAppConfig:
        """Configure the global application.

        Parameters:
            **properties (Dict[str, Any]): A dictionary of additional properties.

        Returns:
            The global application configuration.
        """
        glob_cfg = GlobalAppConfig(**properties)
        if cls._python_config._global_config is None:
            cls._python_config._global_config = glob_cfg
        else:
            cls._python_config._global_config._update(glob_cfg._to_dict())
        cls._compile_configs()
        return cls._applied_config._global_config

    @classmethod
    def check(cls) -> IssueCollector:
        """Check configuration.

        This method logs issue messages and returns an issue collector.

        Returns:
            Collector containing the info, warning and error issues.

        Raises:
            SystemExit: If configuration errors are found, the application
                exits with an error message.
        """
        cls._collector = _Checker._check(cls._applied_config)
        cls.__log_message(cls)
        return cls._collector

    @classmethod
    @_ConfigBlocker._check()
    def _register_default(cls, default_section: Section) -> None:
        if isinstance(default_section, UniqueSection):
            if cls._default_config._unique_sections.get(default_section.name, None):
                cls._default_config._unique_sections[default_section.name]._update(default_section._to_dict())
            else:
                cls._default_config._unique_sections[default_section.name] = default_section
        elif def_sections := cls._default_config._sections.get(default_section.name, None):
            def_sections[default_section.id] = default_section
        else:
            cls._default_config._sections[default_section.name] = {default_section.id: default_section}
        cls._serializer._section_class[default_section.name] = default_section.__class__  # type: ignore
        cls.__json_serializer._section_class[default_section.name] = default_section.__class__  # type: ignore
        cls._compile_configs()

    @classmethod
    @_ConfigBlocker._check()
    def _register(cls, section) -> None:
        if isinstance(section, UniqueSection):
            if cls._python_config._unique_sections.get(section.name, None):
                cls._python_config._unique_sections[section.name]._update(section._to_dict())
            else:
                cls._python_config._unique_sections[section.name] = section
        elif sections := cls._python_config._sections.get(section.name, None):
            if sections.get(section.id, None):
                sections[section.id]._update(section._to_dict())
            else:
                sections[section.id] = section
        else:
            cls._python_config._sections[section.name] = {section.id: section}
        cls._serializer._section_class[section.name] = section.__class__
        cls.__json_serializer._section_class[section.name] = section.__class__
        cls._compile_configs()

    @classmethod
    def _override_env_file(cls) -> None:
        if cfg_filename := os.environ.get(cls._ENVIRONMENT_VARIABLE_NAME_WITH_CONFIG_PATH):
            if not os.path.exists(cfg_filename):
                cls.__logger.error(
                    f"File '{cfg_filename}' provided by environment variable "
                    f"'{cls._ENVIRONMENT_VARIABLE_NAME_WITH_CONFIG_PATH}' does not exist. "
                    f"No configuration will be loaded from environment variable."
                )
                return
            cls.__logger.info(f"Loading configuration provided by environment variable. Filename: '{cfg_filename}'")
            cls._env_file_config = cls._serializer._read(cfg_filename)
            cls.__logger.info(f"Configuration '{cfg_filename}' successfully loaded.")

    @classmethod
    def _compile_configs(cls) -> None:
        Config._override_env_file()
        cls._applied_config._clean()
        if cls._default_config:
            cls._applied_config._update(cls._default_config)
        if cls._python_config:
            cls._applied_config._update(cls._python_config)
        if cls._file_config:
            cls._applied_config._update(cls._file_config)
        if cls._env_file_config:
            cls._applied_config._update(cls._env_file_config)

    @classmethod
    def __log_message(cls, config) -> None:
        for issue in config._collector._warnings:
            cls.__logger.warning(str(issue))
        for issue in config._collector._infos:
            cls.__logger.info(str(issue))
        for issue in config._collector._errors:
            cls.__logger.error(str(issue))
        if len(config._collector._errors) != 0:
            raise SystemExit("Configuration errors found. Please check the error log for more information.")

    @classmethod
    def _to_json(cls, _config: _Config) -> str:
        return cls.__json_serializer._serialize(_config)

    @classmethod
    def _from_json(cls, config_as_str: str) -> _Config:
        return cls.__json_serializer._deserialize(config_as_str)


Config._override_env_file()



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import Any, Dict, Optional

from ..common._config_blocker import _ConfigBlocker
from ..common._template_handler import _TemplateHandler as _tpl


class GlobalAppConfig:
    """Configuration attributes related to the global application."""

    def __init__(self, **properties):
        self._properties = properties

    @property
    def properties(self) -> Dict[str, Any]:
        """A dictionary of additional properties."""
        return {k: _tpl._replace_templates(v) for k, v in self._properties.items()}

    @properties.setter  # type: ignore
    @_ConfigBlocker._check()
    def properties(self, val) -> None:
        self._properties = val

    def __getattr__(self, item: str) -> Optional[Any]:
        return _tpl._replace_templates(self._properties.get(item))

    @classmethod
    def default_config(cls) -> GlobalAppConfig:
        """Return the GlobalAppConfig section used by default.

        Returns:
            The default configuration.
        """
        return GlobalAppConfig()

    def _clean(self):
        self._properties.clear()

    def _to_dict(self):
        as_dict = {}
        as_dict.update(self._properties)
        return as_dict

    @classmethod
    def _from_dict(cls, config_as_dict: Dict[str, Any]):
        config = GlobalAppConfig()
        config._properties = config_as_dict
        return config

    def _update(self, config_as_dict):
        self._properties.update(config_as_dict)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Any, List

from .issue import Issue


class IssueCollector:
    """A collection of configuration issues (instances of class `Issue^`).

    `IssueCollector` is a generic class that collects issues detected during the validation
    process. In particular, an `IssueCollector` is created and returned by the `Config.check()^`
    method. It contains all the collected issues separated by severity (ERROR, WARNING, INFO).
    Each issue is an instance of the class `Issue^` and contains the necessary information to
    understand the issue and help the user to fix it.
    """

    _ERROR_LEVEL = "ERROR"
    _WARNING_LEVEL = "WARNING"
    _INFO_LEVEL = "INFO"

    def __init__(self) -> None:
        self._errors: List[Issue] = []
        self._warnings: List[Issue] = []
        self._infos: List[Issue] = []

    @property
    def all(self) -> List[Issue]:
        """List of all issues collected ordered by decreasing level (ERROR, WARNING and INFO)."""
        return self._errors + self._warnings + self._infos

    @property
    def infos(self) -> List[Issue]:
        """List INFO issues collected."""
        return self._infos

    @property
    def warnings(self) -> List[Issue]:
        """List WARNING issues collected."""
        return self._warnings

    @property
    def errors(self) -> List[Issue]:
        """List of ERROR issues collected."""
        return self._errors

    def _add_error(self, field: str, value: Any, message: str, checker_name: str) -> None:
        self._errors.append(Issue(self._ERROR_LEVEL, field, value, message, checker_name))

    def _add_warning(self, field: str, value: Any, message: str, checker_name: str) -> None:
        self._warnings.append(Issue(self._WARNING_LEVEL, field, value, message, checker_name))

    def _add_info(self, field: str, value: Any, message: str, checker_name: str) -> None:
        self._infos.append(Issue(self._INFO_LEVEL, field, value, message, checker_name))



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import keyword

from ..exceptions.exceptions import InvalidConfigurationId

__INVALID_TAIPY_ID_TERMS = ["CYCLE", "SCENARIO", "SEQUENCE", "TASK", "DATANODE"]


def _validate_id(name: str):
    for invalid__id_term in __INVALID_TAIPY_ID_TERMS:
        if invalid__id_term in name:
            raise InvalidConfigurationId(f"'{name}' is not a valid identifier. '{invalid__id_term}' is restricted.")

    if name.isidentifier() and not keyword.iskeyword(name):
        return name

    raise InvalidConfigurationId(f"'{name}' is not a valid identifier.")



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from importlib.util import find_spec

if find_spec("taipy"):
    if find_spec("taipy.common") and find_spec("taipy.common.config"):
        from taipy.common.config._init import *

    if find_spec("taipy.gui"):
        from taipy.gui._init import *

    if find_spec("taipy.core"):
        from taipy.core._init import *

    if find_spec("taipy.rest"):
        from taipy.rest._init import *

    if find_spec("taipy.gui_core"):
        from taipy.gui_core._init import *

    if find_spec("taipy.enterprise"):
        from taipy.enterprise._init import *

    if find_spec("taipy.designer"):
        from taipy.designer._init import *

    if find_spec("taipy._run"):
        from taipy._run import _run as run



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from ..common._repr_enum import _ReprEnum


class Frequency(_ReprEnum):
    """Frequency of the recurrence of `Cycle^` and `Scenario^` objects.

    This enumeration can have the following values:

    - `DAILY`: Daily frequency, a new cycle is created for each day.
    - `WEEKLY`: Weekly frequency, a new cycle is created for each week (from Monday to Sunday).
    - `MONTHLY`: Monthly frequency, a new cycle is created for each month.
    - `QUARTERLY`: Quarterly frequency, a new cycle is created for each quarter.
    - `YEARLY`: Yearly frequency, a new cycle is created for each year.

    The frequency must be provided in the `ScenarioConfig^`.

    Each recurrent scenario is attached to the cycle corresponding to the creation date and the
    frequency. In other words, each cycle represents an iteration and contains the various scenarios
    created during this iteration.

    For instance, when scenarios have a _MONTHLY_ frequency, one cycle will be created for each
    month (January, February, March, etc.). A new scenario created on February 10th, gets
    attached to the _February_ cycle.
    """

    DAILY = 1
    WEEKLY = 2
    MONTHLY = 3
    QUARTERLY = 4
    YEARLY = 5



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class Issue:
    """An issue detected in the configuration.

    `Issue` is a dataclass that represents an issue detected during the configuration check
    process. It contains the necessary information to understand the issue and help the user to fix
    it.
    """

    level: str
    """Level of the issue among ERROR, WARNING, INFO."""
    field: str
    """Configuration field on which the issue has been detected."""
    value: Any
    """Value of the field on which the issue has been detected."""
    message: str
    """Human readable message to help the user fix the issue."""
    tag: Optional[str]
    """Optional tag to be used to filter issues."""

    def __str__(self) -> str:
        """Return a human-readable string representation of the issue."""
        message = self.message

        if self.value:
            current_value_str = f'"{self.value}"' if isinstance(self.value, str) else f"{self.value}"
            message += f" Current value of property `{self.field}` is {current_value_str}."

        return message



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""The setup script for taipy-common package"""

import json
from pathlib import Path

from setuptools import setup

root_folder = Path(__file__).parent

with open(root_folder / "taipy" / "common" / "version.json") as version_file:
    version = json.load(version_file)
    version_string = f'{version.get("major")}.{version.get("minor")}.{version.get("patch")}'
    if vext := version.get("ext"):
        version_string = f"{version_string}.{vext}"

setup(
    version=version_string,
    install_requires=[r for r in (root_folder / "setup.requirements.txt").read_text("UTF-8").splitlines() if r]
)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import ast
import re
from pathlib import Path
from typing import List

_end_doc = re.compile(r"\"\"\"\s*(#\s*noqa\s*:\s*E501)?\s*\n")


def _get_function_delimiters(initial_line, lines):
    begin = end = initial_line
    while True:
        if lines[begin - 1] == "\n":
            break
        begin -= 1

    if lines[end].endswith("(\n"):
        while ":\n" not in lines[end]:
            end += 1

    if '"""' in lines[end + 1]:
        while True:
            if _end_doc.search(lines[end]):
                break
            end += 1
    return begin, end + 1


def _get_file_lines(filename: str) -> List[str]:
    # Get file lines for later
    with open(filename) as f:
        return f.readlines()


def _get_file_ast(filename: str):
    # Get raw text and build ast
    _config = Path(filename)
    _tree = _config.read_text()
    return ast.parse(_tree)


def _build_base_config_pyi(filename, base_pyi) -> str:
    lines = _get_file_lines(filename)
    tree = _get_file_ast(filename)

    class_lineno = [f.lineno for f in ast.walk(tree) if isinstance(f, ast.ClassDef) and f.name == "Config"]
    begin_class, end_class = _get_function_delimiters(class_lineno[0] - 1, lines)

    base_pyi += "".join(lines[begin_class:end_class])
    functions = [f.lineno for f in ast.walk(tree) if isinstance(f, ast.FunctionDef) and not f.name.startswith("__")]

    for ln in functions:
        begin_line, end_line = _get_function_delimiters(ln - 1, lines)
        base_pyi += "".join(lines[begin_line:end_line])

        base_pyi = __add_docstring(base_pyi, lines, end_line)
        base_pyi += "\n"

    return base_pyi


def __add_docstring(base_pyi, lines, end_line):
    if '"""' not in lines[end_line - 1]:
        base_pyi += '\t\t""""""\n'.replace("\t", "    ")
    return base_pyi


def _build_entity_config_pyi(base_pyi, filename, entity_map) -> str:
    lines = _get_file_lines(filename)
    tree = _get_file_ast(filename)
    functions = {}

    for f in ast.walk(tree):
        if isinstance(f, ast.FunctionDef):
            if "_configure" in f.name and not f.name.startswith("__"):
                functions[f.name] = f.lineno
            elif "_set_default" in f.name and not f.name.startswith("__"):
                functions[f.name] = f.lineno
            elif "_add" in f.name and not f.name.startswith("__"):
                functions[f.name] = f.lineno

    for k, v in functions.items():
        begin_line, end_line = _get_function_delimiters(v - 1, lines)
        try:
            func = "".join(lines[begin_line:end_line])
            func = func if not k.startswith("_") else func.replace(k, entity_map.get(k))
            func = __add_docstring(func, lines, end_line) + "\n"
            base_pyi += func
        except Exception:
            print(f"key={k}")  # noqa: T201
            raise

    return base_pyi


def _generate_entity_and_property_maps(filename):
    entities_map = {}
    property_map = {}
    entity_tree = _get_file_ast(filename)
    functions = [
        f for f in ast.walk(entity_tree) if isinstance(f, ast.Call) and getattr(f.func, "id", "") == "_inject_section"
    ]

    for f in functions:
        entity = ast.unparse(f.args[0])
        entities_map[entity] = {}
        property_map[eval(ast.unparse(f.args[1]))] = entity
        # Remove class name from function map
        text = ast.unparse(f.args[-1]).replace(f"{entity}.", "")
        matches = re.findall(r"\((.*?)\)", text)

        for m in matches:
            v, k = m.replace("'", "").split(",")
            entities_map[entity][k.strip()] = v
    return entities_map, property_map


def _generate_acessors(base_pyi, property_map) -> str:
    for property, cls in property_map.items():
        return_template = f"Dict[str, {cls}]" if property != "job_config" else f"{cls}"
        template = ("\t@_Classproperty\n" + f'\tdef {property}(cls) -> {return_template}:\n\t\t""""""\n').replace(
            "\t", "    "
        )
        base_pyi += template + "\n"
    return base_pyi


def _build_header(filename) -> str:
    _file = Path(filename)
    return _file.read_text() + "\n"


if __name__ == "__main__":
    header_file = "taipy/common/config/stubs/pyi_header.py"
    config_init = Path("taipy/core/config/__init__.py")
    base_config = "taipy/common/config/config.py"

    dn_filename = "taipy/core/config/data_node_config.py"
    job_filename = "taipy/core/config/job_config.py"
    scenario_filename = "taipy/core/config/scenario_config.py"
    task_filename = "taipy/core/config/task_config.py"
    core_filename = "taipy/core/config/core_section.py"

    entities_map, property_map = _generate_entity_and_property_maps(config_init)
    pyi = _build_header(header_file)
    pyi = _build_base_config_pyi(base_config, pyi)
    pyi = _generate_acessors(pyi, property_map)
    pyi = _build_entity_config_pyi(pyi, scenario_filename, entities_map["ScenarioConfig"])
    pyi = _build_entity_config_pyi(pyi, dn_filename, entities_map["DataNodeConfig"])
    pyi = _build_entity_config_pyi(pyi, task_filename, entities_map["TaskConfig"])
    pyi = _build_entity_config_pyi(pyi, job_filename, entities_map["JobConfig"])
    pyi = _build_entity_config_pyi(pyi, core_filename, entities_map["CoreSection"])

    # Remove the final redundant \n
    pyi = pyi[:-1]

    with open("taipy/common/config/config.pyi", "w") as f:
        f.writelines(pyi)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import inspect
import re
import types
from abc import abstractmethod
from datetime import datetime, timedelta
from typing import Any, Dict, Optional

from .._config import _Config
from ..common._template_handler import _TemplateHandler
from ..common._validate_id import _validate_id
from ..common.frequency import Frequency
from ..common.scope import Scope
from ..exceptions.exceptions import LoadingError
from ..global_app.global_app_config import GlobalAppConfig
from ..section import Section
from ..unique_section import UniqueSection


class _BaseSerializer(object):
    """Base serializer class for taipy configuration."""

    _GLOBAL_NODE_NAME = "TAIPY"
    _section_class = {_GLOBAL_NODE_NAME: GlobalAppConfig}

    @classmethod
    @abstractmethod
    def _write(cls, configuration: _Config, filename: str):
        raise NotImplementedError

    @classmethod
    def _str(cls, configuration: _Config):
        config_as_dict = {cls._GLOBAL_NODE_NAME: configuration._global_config._to_dict()}
        for u_sect_name, u_sect in configuration._unique_sections.items():
            config_as_dict[u_sect_name] = u_sect._to_dict()
        for sect_name, sections in configuration._sections.items():
            config_as_dict[sect_name] = cls._to_dict(sections)
        return cls._stringify(config_as_dict)

    @classmethod
    def _to_dict(cls, sections: Dict[str, Any]):
        return {section_id: section._to_dict() for section_id, section in sections.items()}

    @classmethod
    def _stringify(cls, as_dict):
        if as_dict is None:
            return None
        if isinstance(as_dict, Section):
            return f"{as_dict.id}:SECTION"
        if isinstance(as_dict, Scope):
            return f"{as_dict.name}:SCOPE"
        if isinstance(as_dict, Frequency):
            return f"{as_dict.name}:FREQUENCY"
        if isinstance(as_dict, bool):
            return f"{str(as_dict)}:bool"
        if isinstance(as_dict, int):
            return f"{str(as_dict)}:int"
        if isinstance(as_dict, float):
            return f"{str(as_dict)}:float"
        if isinstance(as_dict, datetime):
            return f"{as_dict.isoformat()}:datetime"
        if isinstance(as_dict, timedelta):
            return f"{cls._timedelta_to_str(as_dict)}:timedelta"
        if inspect.isfunction(as_dict) or isinstance(as_dict, types.BuiltinFunctionType):
            return f"{as_dict.__module__}.{as_dict.__name__}:function"
        if inspect.isclass(as_dict):
            return f"{as_dict.__module__}.{as_dict.__qualname__}:class"
        if isinstance(as_dict, dict):
            return {str(key): cls._stringify(val) for key, val in as_dict.items()}
        if isinstance(as_dict, list):
            return [cls._stringify(val) for val in as_dict]
        if isinstance(as_dict, tuple):
            return [cls._stringify(val) for val in as_dict]
        if isinstance(as_dict, set):
            return [cls._stringify(val) for val in as_dict]
        return as_dict

    @staticmethod
    def _extract_node(config_as_dict, cls_config, node, config: Optional[Any]) -> Dict[str, Section]:
        res = {}
        for key, value in config_as_dict.get(node, {}).items():  # my_task, {input=[], output=[my_data_node], ...}
            key = _validate_id(key)
            res[key] = cls_config._from_dict(value, key, config)  # if config is None else cls_config._from_dict(key,
            # value, config)
        return res

    @classmethod
    def _from_dict(cls, as_dict) -> _Config:
        config = _Config()
        config._global_config = GlobalAppConfig._from_dict(as_dict.get(cls._GLOBAL_NODE_NAME, {}))
        for section_name, sect_as_dict in as_dict.items():
            if section_class := cls._section_class.get(section_name, None):
                if issubclass(section_class, UniqueSection):
                    config._unique_sections[section_name] = section_class._from_dict(  # type: ignore
                        sect_as_dict, None, None
                    )
                elif issubclass(section_class, Section):
                    config._sections[section_name] = cls._extract_node(as_dict, section_class, section_name, config)
        return config

    @classmethod
    def _pythonify(cls, val):
        match = re.fullmatch(_TemplateHandler._PATTERN, str(val))
        if not match:
            if isinstance(val, str):
                TYPE_PATTERN = (
                    r"^(.+):(\bbool\b|\bstr\b|\bint\b|\bfloat\b|\bdatetime\b||\btimedelta\b|"
                    r"\bfunction\b|\bclass\b|\bSCOPE\b|\bFREQUENCY\b|\bSECTION\b)?$"
                )
                if match := re.fullmatch(TYPE_PATTERN, str(val)):
                    actual_val = match.group(1)
                    dynamic_type = match.group(2)
                    if dynamic_type == "SECTION":
                        return actual_val
                    if dynamic_type == "FREQUENCY":
                        return Frequency[actual_val]
                    if dynamic_type == "SCOPE":
                        return Scope[actual_val]
                    if dynamic_type == "bool":
                        return _TemplateHandler._to_bool(actual_val)
                    elif dynamic_type == "int":
                        return _TemplateHandler._to_int(actual_val)
                    elif dynamic_type == "float":
                        return _TemplateHandler._to_float(actual_val)
                    elif dynamic_type == "datetime":
                        return _TemplateHandler._to_datetime(actual_val)
                    elif dynamic_type == "timedelta":
                        return _TemplateHandler._to_timedelta(actual_val)
                    elif dynamic_type == "function":
                        return _TemplateHandler._to_function(actual_val)
                    elif dynamic_type == "class":
                        return _TemplateHandler._to_class(actual_val)
                    elif dynamic_type == "str":
                        return actual_val
                    else:
                        error_msg = f"Error loading toml configuration at {val}. {dynamic_type} type is not supported."
                        raise LoadingError(error_msg)
            if isinstance(val, dict):
                return {str(k): cls._pythonify(v) for k, v in val.items()}
            if isinstance(val, list):
                return [cls._pythonify(v) for v in val]
        return val

    @classmethod
    def _timedelta_to_str(cls, obj: timedelta) -> str:
        total_seconds = obj.total_seconds()
        return (
            f"{int(total_seconds // 86400)}d"
            f"{int(total_seconds % 86400 // 3600)}h"
            f"{int(total_seconds % 3600 // 60)}m"
            f"{int(total_seconds % 60)}s"
        )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
import json
from datetime import timedelta
from typing import Any, Callable, Dict, List, Optional, Union

from taipy.common.config._config import _Config
from taipy.core.config import CoreSection, DataNodeConfig, JobConfig, ScenarioConfig, TaskConfig

from .checker.issue_collector import IssueCollector
from .common._classproperty import _Classproperty
from .common._config_blocker import _ConfigBlocker
from .common.frequency import Frequency
from .common.scope import Scope
from .global_app.global_app_config import GlobalAppConfig
from .section import Section
from .unique_section import UniqueSection



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import pathlib
import sys
from typing import Dict, Optional

from cookiecutter.exceptions import OutputDirExistsException
from cookiecutter.main import cookiecutter

import taipy

from ._base_cli._abstract_cli import _AbstractCLI
from ._base_cli._taipy_parser import _TaipyParser


class _CreateCLI(_AbstractCLI):
    _template_map: Dict[str, str] = {}

    _COMMAND_NAME = "create"
    _ARGUMENTS = ["--application"]

    @classmethod
    def generate_template_map(cls, template_path: Optional[pathlib.Path] = None):
        if not template_path:
            template_path = pathlib.Path(taipy.__file__).parent.resolve() / "templates"

        # Update the template map with the new templates but do not override the existing ones
        cls._template_map.update(
            {
                str(x.name): str(x)
                for x in template_path.iterdir()
                if x.is_dir() and not x.name.startswith("_") and x.name not in cls._template_map
            }
        )

    @classmethod
    def create_parser(cls):
        create_parser = _TaipyParser._add_subparser(
            cls._COMMAND_NAME,
            help="Create a new Taipy application using pre-defined templates.",
        )
        create_parser.add_argument(
            "--application",
            choices=list(cls._template_map.keys()),
            default="default",
            help="The template used to create the new Taipy application.",
        )

    @classmethod
    def handle_command(cls):
        args = cls._parse_arguments()
        if not args:
            return
        try:
            cookiecutter(cls._template_map[args.application])
        except OutputDirExistsException as err:
            error_msg = f"{str(err)}. Please remove the existing directory or provide a new folder name."
            print(error_msg)  # noqa: T201
            sys.exit(1)
        sys.exit(0)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import re
from collections import UserDict
from datetime import datetime, timedelta
from importlib import import_module
from operator import attrgetter
from pydoc import locate

from ..exceptions.exceptions import InconsistentEnvVariableError, MissingEnvVariableError
from .frequency import Frequency
from .scope import Scope


class _TemplateHandler:
    """Factory to handle actions related to config value templating."""

    _PATTERN = r"^ENV\[([a-zA-Z_]\w*)\](:(\bbool\b|\bstr\b|\bfloat\b|\bint\b))?$"

    @classmethod
    def _replace_templates(cls, template, type=str, required=True, default=None):
        if isinstance(template, tuple):
            return tuple(cls._replace_template(item, type, required, default) for item in template)
        if isinstance(template, list):
            return [cls._replace_template(item, type, required, default) for item in template]
        if isinstance(template, dict):
            return {str(k): cls._replace_template(v, type, required, default) for k, v in template.items()}
        if isinstance(template, UserDict):
            return {str(k): cls._replace_template(v, type, required, default) for k, v in template.items()}
        return cls._replace_template(template, type, required, default)

    @classmethod
    def _replace_template(cls, template, type, required, default):
        if "ENV" not in str(template):
            return template
        if match := re.fullmatch(cls._PATTERN, str(template)):
            var = match.group(1)
            dynamic_type = match.group(3)
            val = os.environ.get(var)
            if val is None:
                if required:
                    raise MissingEnvVariableError(f"Environment variable {var} is not set.")
                return default
            if type is bool:
                return cls._to_bool(val)
            elif type is int:
                return cls._to_int(val)
            elif type is float:
                return cls._to_float(val)
            elif type is Scope:
                return cls._to_scope(val)
            elif type is Frequency:
                return cls._to_frequency(val)
            else:
                if dynamic_type == "bool":
                    return cls._to_bool(val)
                elif dynamic_type == "int":
                    return cls._to_int(val)
                elif dynamic_type == "float":
                    return cls._to_float(val)
                return val
        return template

    @staticmethod
    def _to_bool(val: str) -> bool:
        possible_values = ["true", "false"]
        if str.lower(val) not in possible_values:
            raise InconsistentEnvVariableError("{val} is not a Boolean.")
        return str.lower(val) == "true" or str.lower(val) != "false"

    @staticmethod
    def _to_int(val: str) -> int:
        try:
            return int(val)
        except ValueError:
            raise InconsistentEnvVariableError(f"{val} is not an integer.") from None

    @staticmethod
    def _to_float(val: str) -> float:
        try:
            return float(val)
        except ValueError:
            raise InconsistentEnvVariableError(f"{val} is not a float.") from None

    @staticmethod
    def _to_datetime(val: str) -> datetime:
        try:
            return datetime.fromisoformat(val)
        except ValueError:
            raise InconsistentEnvVariableError(f"{val} is not a valid datetime.") from None

    @staticmethod
    def _to_timedelta(val: str) -> timedelta:
        """
        Parse a time string e.g. (2h13m) into a timedelta object.

        :param timedelta_str: A string identifying a duration.  (eg. 2h13m)
        :return datetime.timedelta: A datetime.timedelta object
        """
        regex = re.compile(
            r"^((?P<days>[\.\d]+?)d)? *"
            r"((?P<hours>[\.\d]+?)h)? *"
            r"((?P<minutes>[\.\d]+?)m)? *"
            r"((?P<seconds>[\.\d]+?)s)?$"
        )
        parts = regex.match(val)
        if not parts:
            raise InconsistentEnvVariableError(f"{val} is not a valid timedelta.")
        time_params = {name: float(param) for name, param in parts.groupdict().items() if param}
        return timedelta(**time_params)  # type: ignore

    @staticmethod
    def _to_scope(val: str) -> Scope:
        try:
            return Scope[str.upper(val)]
        except Exception:
            raise InconsistentEnvVariableError(f"{val} is not a valid scope.") from None

    @staticmethod
    def _to_frequency(val: str) -> Frequency:
        try:
            return Frequency[str.upper(val)]
        except Exception:
            raise InconsistentEnvVariableError(f"{val} is not a valid frequency.") from None

    @staticmethod
    def _to_function(val: str):
        module_name, fct_name = val.rsplit(".", 1)
        try:
            module = import_module(module_name)
            return attrgetter(fct_name)(module)
        except Exception:
            raise InconsistentEnvVariableError(f"{val} is not a valid function.") from None

    @staticmethod
    def _to_class(val: str):
        try:
            return locate(val)
        except Exception:
            raise InconsistentEnvVariableError(f"{val} is not a valid class.") from None



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import List, Type

from ..common.typing import ConfigCheckerType
from .issue_collector import IssueCollector


class _Checker:
    """Holds the various checkers to perform on the config."""

    _checkers: List = []

    @classmethod
    def _check(cls, _applied_config):
        collector = IssueCollector()
        for checker in cls._checkers:
            checker(_applied_config, collector)._check()
        return collector

    @classmethod
    def add_checker(cls, checker_class: Type[ConfigCheckerType]):
        cls._checkers.append(checker_class)
